"""Module provider for Joker.com"""
# The Joker.com API is well documented.
# Several specificities compared to classic REST API in other providers:
#   - everything is done with GET requests: all parameters (including actual records data
#     and authentication data) are passed in the URLs,
#   - DNS zones are represented in a "strongly" formatted flat format (not JSON or XML) that may
#     remind a zone definition in bind9,
#   - every operation requires to pass the entire updated zone through the API, so care must be
#     taken to not alter unexpectedly other entries during the create/update/delete operations,
#   - all headers and data are contained in the response body; then this kind of body is composed
#     of several lines of type key: value containing the headers (including errors), then a blank
#     line makes the separation with the data itself (see _process_response for the body parsing).
import binascii
import json
import logging
import re

import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["ns.joker.com"]

API_BASE_URL = "https://dmapi.joker.com/request"
DNSZONE_ENTRY_PATTERN = re.compile(
    r"^(.+?)\s+(\w+)\s+(.+?)\s+((?:\\\"(?:.+?)\\\")|(?:.+?))\s+(\d+)(?:\s+(\d+)\s+(\d+)\s+(.+?)|\s+(.+?)|)$"
)


def provider_parser(subparser):
    """Generate a subparser for Joker"""
    subparser.description = """
The Joker.com provider requires a valid token for authentication.
You can create one in the section 'Manage Joker.com API access keys' of 'My Profile' in your Joker.com account.
"""
    subparser.add_argument(
        "--auth-token",
        help="specify the API Key to connect to the Joker.com API",
    )


class _Response:
    def __init__(self, headers, data):
        self.headers = headers
        self.data = data

    def __str__(self):
        return json.dumps({"headers": self.headers, "data": self.data})


class Provider(BaseProvider):
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self._session_id = None

    def _authenticate(self):
        auth_token = self._get_provider_option("auth_token")

        response = requests.post(API_BASE_URL + "/login", data={"api-key": auth_token})

        result = _process_response(response)
        self._session_id = result.headers["Auth-Sid"]

        response = self._post(
            "/query-domain-list", data={"pattern": self.domain, "showstatus": 1}
        )

        if not response.data:
            raise AuthenticationError(
                f"Domain {self.domain} is not registered with this account."
            )

        data = response.data[0]
        items = data.split(" ")
        domain_status = items[2].split(",")

        if len(set(domain_status).difference(["production", "lock", "autorenew"])) > 0:
            raise AuthenticationError(
                f"Current status for domain {self.domain} is: {items[2]}"
            )

        self.domain_id = self.domain

    def _request(self, action="POST", url="/", data=None, query_params=None):
        if not data:
            data = {}

        data["auth-sid"] = self._session_id

        response = requests.post(API_BASE_URL + url, data=data)
        return _process_response(response)

    def _list_records(self, rtype=None, name=None, content=None):
        response = self._post("/dns-zone-get", data={"domain": self.domain_id})
        zone_data = _extract_zonedata(response.data)

        zone_data = [entry for entry in zone_data if entry["type"]]

        if rtype:
            zone_data = [entry for entry in zone_data if entry["type"] == rtype]
        if name:
            zone_data = [
                entry
                for entry in zone_data
                if self._full_name(entry["label"]) == self._full_name(name)
            ]
        if content:
            zone_data = [
                entry for entry in zone_data if entry["target"] == content.strip()
            ]

        records = [
            {
                "type": entry["type"],
                "name": self._full_name(entry["label"]),
                "ttl": entry["ttl"],
                "content": entry["target"],
                "id": self._identifier(entry),
            }
            for entry in zone_data
        ]

        LOGGER.debug("list_records: %s", records)

        return records

    def _create_record(self, rtype, name, content):
        if not rtype or not name or not content:
            raise Exception(
                "Error, rtype, name and content are mandatory to create a record."
            )

        response = self._post("/dns-zone-get", data={"domain": self.domain_id})
        zonedata = _extract_zonedata(response.data)

        new_entry = {
            "label": self._relative_name(name),
            "type": rtype,
            "pri": 0,
            "target": content,
            "ttl": self._get_lexicon_option("ttl"),
            "valid-from": None,
            "valid-to": None,
            "parameters": None,
        }

        if any(
            entry
            for entry in zonedata
            if self._identifier(new_entry) == self._identifier(entry)
        ):
            LOGGER.debug(
                "create_record (ignored, duplicate): %s", self._identifier(new_entry)
            )
            return True

        zonedata.append(new_entry)

        self._apply_zonedata(zonedata)

        return True

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier and not (rtype and name):
            raise Exception(
                "Error, either identifier or rtype + name are mandatory to update a record."
            )

        response = self._post("/dns-zone-get", data={"domain": self.domain_id})
        zonedata = _extract_zonedata(response.data)

        selector_info = (
            f"identifier={identifier}" if identifier else f"type={rtype},name={name}"
        )

        if identifier:
            to_update = [
                entry for entry in zonedata if self._identifier(entry) == identifier
            ]
        else:
            to_update = [
                entry
                for entry in zonedata
                if entry["type"] == rtype
                and self._full_name(entry["label"]) == self._full_name(name)
            ]

        if not to_update:
            raise Exception(f"Error, could not find a record for {selector_info}.")

        if len(to_update) > 1:
            raise Exception(
                f"Error, found more than one record for {selector_info}. "
                "Please use an identifier to select one explicitly."
            )

        to_update[0].update(
            {
                "type": rtype if rtype else to_update[0]["type"],
                "label": self._relative_name(name) if name else to_update[0]["label"],
                "target": content if content else to_update[0]["content"],
                "ttl": self._get_lexicon_option("ttl"),
            }
        )

        self._apply_zonedata(zonedata)

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier and not rtype:
            raise Exception(
                "Error, either rtype or identifier are mandatory to delete a record."
            )

        response = self._post("/dns-zone-get", data={"domain": self.domain_id})
        zonedata = _extract_zonedata(response.data)

        if identifier:
            zonedata = [
                entry for entry in zonedata if self._identifier(entry) != identifier
            ]
        else:
            zonedata = [
                entry
                for entry in zonedata
                if entry["type"] is None
                or self._identifier(entry)
                != self._identifier(
                    {
                        "type": rtype,
                        "label": name if name else entry["label"],
                        "target": content if content else entry["target"],
                    }
                )
            ]

        self._apply_zonedata(zonedata)

        return True

    def _apply_zonedata(self, zonedata):
        data = []
        for entry in zonedata:
            if "raw" in entry:
                data.append(entry["raw"])
            else:
                # TXT entries always require content to be quoted
                target = (
                    f'"{entry["target"]}"'
                    if entry["type"] == "TXT"
                    else entry["target"]
                )
                line = f"{entry['label']} {entry['type']} {entry['pri']} {target} {entry['ttl']}"

                if entry["valid-from"] is not None and entry["valid-to"] is not None:
                    line = f"{line} {entry['valid-from']} {entry['valid-to']}"

                if entry["parameters"] is not None:
                    line = f"{line} {entry['parameters']}"

                data.append(line)

        self._post(
            "/dns-zone-put",
            data={"domain": self.domain_id, "zone": "\n".join(data)},
        )

    def _identifier(self, record):
        if "raw" in record:
            return None

        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
        digest.update(
            ("name=" + self._full_name(record.get("label", "")) + ",").encode("utf-8")
        )
        digest.update(("content=" + record.get("target", "") + ",").encode("utf-8"))

        return binascii.hexlify(digest.finalize()).decode("utf-8")[0:7]

    def _relative_name(self, record_name):
        if record_name == self.domain_id:
            return "@"

        return super(Provider, self)._relative_name(record_name)

    def _full_name(self, record_name):
        if record_name == "@":
            return self.domain_id

        return super(Provider, self)._full_name(record_name)


def _extract_zonedata(data):
    processed = []

    for entry in data:
        match = DNSZONE_ENTRY_PATTERN.match(entry)

        if match:
            extracted = {
                "label": match.group(1),
                "type": match.group(2),
                "pri": match.group(3),
                "target": match.group(4),
                "ttl": int(match.group(5)),
            }

            try:
                extracted.update(
                    {
                        "parameters": match.group(9),
                        "valid-from": None,
                        "valid-to": None,
                    }
                )
            except IndexError:
                try:
                    extracted.update(
                        {
                            "parameters": match.group(8),
                            "valid-from": int(match.group(7)),
                            "valid-to": int(match.group(6)),
                        }
                    )
                except IndexError:
                    pass

            if extracted["type"] == "TXT":
                extracted["target"] = re.sub(
                    r'"(.*)"', r"\1", extracted["target"]
                ).strip()

            processed.append(extracted)
        else:
            processed.append(
                {
                    "raw": entry,
                    "type": None,
                }
            )

    return processed


def _process_response(response):
    response.raise_for_status()

    headers = {}
    body = []
    feed_headers = True

    data = response.text
    for line in data.split("\n"):
        if not line:
            feed_headers = False
            continue

        if feed_headers:
            items = line.split(":")
            headers[items[0]] = items[1].lstrip()
        else:
            body.append(line)

    processed_response = _Response(headers, body)

    if headers["Status-Code"] != "0":
        if body and body[0] == "API key is invalid":
            raise requests.exceptions.HTTPError(
                f"{headers['Status-Code']} Error: API key is invalid for url: {response.url}",
                response=processed_response,
            )
        raise requests.exceptions.HTTPError(
            f"{headers['Status-Code']} Error: {headers['Status-Text']} ({headers['Error']}) "
            f"for url: {response.url}",
            response=processed_response,
        )

    return processed_response



"""Module provider for EasyDNS"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["easydns.net"]


def provider_parser(subparser):
    """Configure provider parser for EasyDNS"""
    subparser.add_argument(
        "--auth-username", help="specify username for authentication"
    )
    subparser.add_argument("--auth-token", help="specify token for authentication")


class Provider(BaseProvider):
    """Provider class for EasyDNS"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = (
            self._get_provider_option("api_endpoint") or "https://rest.easydns.net"
        )

    def _authenticate(self):

        payload = self._get(f"/domain/{self.domain}")

        if payload["data"]["exists"] == "N":
            raise AuthenticationError("No domain found")

        self.domain_id = payload["data"]["id"]

    # Create record. If record already exists with the same content, do nothing'

    def _create_record(self, rtype, name, content):
        record = {
            "type": rtype,
            "domain": self.domain_id,
            "host": self._relative_name(name),
            "ttl": self._get_lexicon_option("ttl"),
            "prio": 0,
            "rdata": content,
        }
        try:
            self._put(f"/zones/records/add/{self.domain_id}/{rtype}", record)
        except requests.exceptions.HTTPError as error:
            # FIXME: adferrand 06/01/2019: Broken provider needs fixes.
            # In fact, this except block will silently hide every HTTP error, as an except
            # block without raise statement will implicitly hide the exception.
            # In reality, tests of Easy DNS are failing...
            # So this provider needs to be corrected.
            if error.response.status_code == 400:
                pass
            # http 400 is ok here, because the record probably already exists
        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get(f"/zones/records/all/{self.domain_id}")
        records = []
        for record in payload["data"]:
            processed_record = {
                "type": record["type"],
                "name": f"{record['host']}.{record['domain']}",
                "ttl": record["ttl"],
                "content": record["rdata"],
                "id": record["id"],
            }
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        data = {"ttl": self._get_lexicon_option("ttl")}
        if rtype:
            data["type"] = rtype
        if name:
            data["host"] = self._relative_name(name)
        if content:
            data["rdata"] = content

        self._post(f"/zones/records/{identifier}", data)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/zones/records/{self.domain_id}/{record_id}")

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        query_params["format"] = "json"
        query_params["_user"] = self._get_provider_option("auth_username")
        query_params["_key"] = self._get_provider_option("auth_token")
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Module provider for memset"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["memset.com"]


def provider_parser(subparser):
    """Configure provider parser for memset"""
    subparser.add_argument("--auth-token", help="specify API key for authentication")


class Provider(BaseProvider):
    """Provider class for memset"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.memset.com/v1/json"

    def _authenticate(self):
        payload = self._get("/dns.zone_domain_info", {"domain": self.domain})
        if not payload["zone_id"]:
            raise AuthenticationError("No domain found")
        self.domain_id = payload["zone_id"]

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        data = {"type": rtype, "record": self._relative_name(name), "address": content}
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")
        data["zone_id"] = self.domain_id
        check_exists = self._list_records(rtype=rtype, name=name, content=content)
        if not check_exists:
            payload = self._get("/dns.zone_record_create", data)
            if payload["id"]:
                self._get("/dns.reload")
                LOGGER.debug("create_record: %s", payload["id"])
                return payload["id"]
        return check_exists

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get("/dns.zone_info", {"id": self.domain_id})
        records = []
        for record in payload["records"]:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["record"]),
                "ttl": record["ttl"],
                "content": record["address"],
                "id": record["id"],
            }
            if name:
                name = self._full_name(name)
            if processed_record["type"] == rtype:
                if name is not None and content is not None:
                    if (
                        processed_record["name"] == name
                        and processed_record["content"] == content
                    ):
                        records.append(processed_record)
                elif name is not None and content is None:
                    if processed_record["name"] == name:
                        records.append(processed_record)
                elif name is None and content is not None:
                    if processed_record["content"] == content:
                        records.append(processed_record)
                else:
                    records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        data = {}
        if not identifier:
            records = self._list_records(rtype, self._relative_name(name))
            if len(records) == 1:
                identifier = records[0]["id"]
            else:
                raise Exception("Record identifier could not be found.")
        if rtype:
            data["type"] = rtype
        if name:
            data["record"] = self._relative_name(name)
        if content:
            data["address"] = content
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")
        data["id"] = identifier
        data["zone_id"] = self.domain_id

        payload = self._get("/dns.zone_record_update", data)
        if payload["id"]:
            self._get("/dns.reload")
            LOGGER.debug("update_record: %s", payload["id"])
            return payload["id"]
        return False

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, self._relative_name(name), content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        record_id = None
        for record_id in delete_record_id:
            self._get("/dns.zone_record_delete", {"id": record_id})

        if record_id:
            self._get("/dns.reload")

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            auth=(self._get_provider_option("auth_token"), "x"),
            headers={"Content-Type": "application/json"},
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Module provider for Dinahosting"""
# import re
import hashlib
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["dinahosting.com"]

# These record types are either not supported by the API or are non-compliant
UNSUPPORTED_TYPES = ["MX", "SRV", "NS", "SOA", "LOC"]

# The API returns the record's content in any of these keys
CONTENT_KEYS = ["ip", "address", "text", "destinationHostname"]

# The API returns the record's name in any of these keys
NAME_KEYS = ["host", "hostname"]

# API response codes
RC_SUCCESS = 1000
RC_SUCCESS_PENDING = 1001
RC_UNKNOWN_COMMAND = 2000
RC_COMMAND_SYNTAX_ERROR = 2001
RC_COMMAND_USE_ERROR = 2002
RC_PARAM_MISSING = 2003
RC_PARAM_VALUE_RANGE = 2004
RC_PARAM_VALUE_SYNTAX = 2005
RC_AUTH_ERROR_USER = 2200
RC_AUTH_ERROR_OBJECT = 2201
RC_OBJECT_EXISTS = 2302
RC_OBJECT_NOT_EXISTS = 2303
RC_COMMAND_FAILED = 2400
RC_COMMAND_FAILED_FATAL = 2500
RC_COMMAND_TIMEOUT = 2501


def provider_parser(subparser):
    """Return the parser for this provider"""
    subparser.add_argument(
        "--auth-username", help="specify username for authentication"
    )
    subparser.add_argument(
        "--auth-password", help="specify password for authentication"
    )


class Provider(BaseProvider):
    """Provider class for Dinahosting"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://dinahosting.com/special/api.php"

    def _authenticate(self):
        data = {"command": "Services_GetDomains"}
        payload = self._post("", data)

        if not payload["data"]:
            raise AuthenticationError("No domain found")

        for data in payload["data"]:
            if data["domain"] == self.domain:
                break
        else:
            raise AuthenticationError("Requested domain is not among the owned domains")

        self.domain_id = self.domain

    def _create_record(self, rtype, name, content):
        Provider._check_unsupported_type(rtype)

        rel_name = self._relative_name(name)
        data = {"domain": self.domain, "hostname": rel_name}

        if rtype == "A":
            data["command"] = "Domain_Zone_AddTypeA"
            data["ip"] = content
        elif rtype == "AAAA":
            data["command"] = "Domain_Zone_AddTypeAAAA"
            data["ip"] = content
        elif rtype == "CNAME":
            data["command"] = "Domain_Zone_AddTypeCname"
            data["destinationHostname"] = content
        elif rtype == "TXT":
            data["command"] = "Domain_Zone_AddTypeTXT"
            data["text"] = content

        try:
            self._post("", data)
        except requests.exceptions.HTTPError as err:
            already_exists = (
                err.response.json()["responseCode"] == RC_OBJECT_EXISTS
            ) or (err.response.json()["responseCode"] == RC_COMMAND_FAILED)
            if not already_exists:
                raise

        LOGGER.debug("create_record: %s", True)
        return True

    def _list_records(self, rtype=None, name=None, content=None):
        if rtype:
            formatted_rtype = "Cname" if rtype == "CNAME" else rtype
            data = {
                "command": f"Domain_Zone_GetType{formatted_rtype}",
                "domain": self.domain,
            }
        else:
            data = {"command": "Domain_Zone_GetAll", "domain": self.domain}

        payload = self._post("", data)

        records = []
        for record in payload["data"]:
            precord = self._parse_record(record)
            ptype = rtype if rtype else precord["type"]

            if name and name not in precord["name_variants"]:
                continue

            if content and content != precord["content"]:
                continue

            if rtype and precord["type"] and rtype != precord["type"]:
                continue

            processed_record = {
                "type": ptype,
                "name": precord["name_variants"][0],
                "content": precord["content"],
                "ttl": None,  # The API does not expose the record's ttl
            }
            processed_record["id"] = Provider._identifier(processed_record)
            records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier:
            records = self._list_records()
            matching_records = [
                record for record in records if record["id"] == identifier
            ]
        else:
            matching_records = self._list_records(rtype, name)

        # Check if record exists, otherwise raise an error
        if not matching_records:
            raise Exception("No matching records found. Cannot update.")
        if len(matching_records) > 1:
            raise Exception("Multiple matching records found. Cannot update.")

        record = matching_records[0]
        # The API does not offer update functions for most of the record types
        delete_result = self._delete_record(
            identifier, record["type"], record["name"], record["content"]
        )
        create_result = self._create_record(rtype, name, content)

        LOGGER.debug("update_record: %s", (delete_result and create_result))
        return delete_result and create_result

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier:
            records = self._list_records()
            delete_records = [
                record for record in records if record["id"] == identifier
            ]
        else:
            delete_records = self._list_records(rtype, name, content)

        for delete_record in delete_records:
            ptype = rtype if rtype else delete_record["type"]
            Provider._check_unsupported_type(ptype)

            rel_name = self._relative_name(delete_record["name"])
            data = {"domain": self.domain, "hostname": rel_name}

            if ptype == "A":
                data["command"] = "Domain_Zone_DeleteTypeA"
                data["ip"] = delete_record["content"]
            elif ptype == "AAAA":
                data["command"] = "Domain_Zone_DeleteTypeAAAA"
                data["ip"] = delete_record["content"]
            elif ptype == "CNAME":
                data["command"] = "Domain_Zone_DeleteTypeCname"
            elif ptype == "TXT":
                data["command"] = "Domain_Zone_DeleteTypeTXT"
                data["text"] = delete_record["content"]

            try:

                self._post("", data)
            except requests.exceptions.HTTPError as err:
                not_exists = err.response.json()["responseCode"] == RC_OBJECT_NOT_EXISTS
                if not not_exists:
                    raise

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        # All requests are POST. Query parameters are not used.
        query_params = {}

        if data is None:
            data = {}
        data["responseType"] = "Json"

        username = self._get_provider_option("auth_username")
        password = self._get_provider_option("auth_password")
        response = requests.request(
            action,
            self.api_endpoint + url,
            auth=(username, password),
            params=query_params,
            data=data,
        )
        # If the request fails for any reason, throw an error.
        response.raise_for_status()

        rjson = response.json()
        if int(rjson["responseCode"]) != RC_SUCCESS:
            api_error_message = "%s API Error for url: %s" % (
                rjson["responseCode"],
                response.url,
            )
            raise requests.exceptions.HTTPError(api_error_message, response=response)
        return rjson

    def _parse_record(self, record):
        parsed_record = {}
        parsed_record["type"] = Provider._parse_record_type(record)
        parsed_record["content"] = Provider._find_matching_record_value(
            CONTENT_KEYS, record
        )
        raw_name = Provider._find_matching_record_value(NAME_KEYS, record)
        fqdn_name = self._fqdn_name(raw_name)
        full_name = self._full_name(raw_name)
        parsed_record["name_variants"] = [full_name, fqdn_name, raw_name]

        return parsed_record

    @staticmethod
    def _identifier(record):
        md5 = hashlib.md5()
        md5.update((record.get("type", "")).encode("utf-8"))
        md5.update((record.get("name", "")).encode("utf-8"))
        md5.update((record.get("data", "")).encode("utf-8"))
        return md5.hexdigest()

    @staticmethod
    def _find_matching_record_value(keys, record):
        if record is None or keys is None:
            return None

        matched_value = None
        for key in keys:
            if key in record:
                matched_value = record[key]
                break
        return matched_value

    @staticmethod
    def _parse_record_type(record):
        if not record or "type" not in record:
            return None
        if record["type"].startswith("MX"):
            return "MX"
        return record["type"]

    @staticmethod
    def _check_unsupported_type(rtype):
        if rtype in UNSUPPORTED_TYPES:
            raise Exception(f"Record type {rtype} is not supported by the API")



"""Module provider for Glesys"""
import json

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

NAMESERVER_DOMAINS = ["glesys.com"]


def provider_parser(subparser):
    """Generate a subparser for Glesys"""
    subparser.add_argument("--auth-username", help="specify username (CL12345)")
    subparser.add_argument("--auth-token", help="specify API key")


class Provider(BaseProvider):
    """Provider class for Glesys"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.glesys.com"

    def _authenticate(self):
        payload = self._get("/domain/list")
        domains = payload["response"]["domains"]
        for record in domains:
            if record["domainname"] == self.domain:
                # Domain records do not have any id.
                # Since domain_id cannot be None, use domain name as id instead.
                self.domain_id = record["domainname"]
                break
        else:
            raise AuthenticationError("No domain found")

    # Create record. If record already exists with the same content, do nothing.
    def _create_record(self, rtype, name, content):
        existing = self.list_records(rtype, name, content)
        if existing:
            # Already exists, do nothing.
            return True

        request_data = {
            "domainname": self.domain,
            "host": name,
            "type": rtype,
            "data": content,
        }
        self._addttl(request_data)

        self._post("/domain/addrecord", data=request_data)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        request_data = {"domainname": self.domain}
        payload = self._post("/domain/listrecords", data=request_data)

        # Convert from Glesys record structure to Lexicon structure.
        processed_records = [
            self._glesysrecord2lexiconrecord(r) for r in payload["response"]["records"]
        ]

        if rtype:
            processed_records = [
                record for record in processed_records if record["type"] == rtype
            ]
        if name:
            processed_records = [
                record
                for record in processed_records
                if record["name"] == self._full_name(name)
            ]
        if content:
            processed_records = [
                record
                for record in processed_records
                if record["content"].lower() == content.lower()
            ]

        return processed_records

    # Update a record. Identifier must be specified.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        request_data = {"recordid": identifier}
        if name:
            request_data["host"] = name
        if rtype:
            request_data["type"] = rtype
        if content:
            request_data["data"] = content

        self._addttl(request_data)
        self._post("/domain/updaterecord", data=request_data)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        for record_id in delete_record_id:
            request_data = {"recordid": record_id}
            self._post("/domain/deleterecord", data=request_data)

        return True

    # Helpers.
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        query_params["format"] = "json"
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        credentials = (
            self._get_provider_option("auth_username"),
            self._get_provider_option("auth_token"),
        )
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
            auth=credentials,
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    # Adds TTL parameter if passed as argument to lexicon.
    def _addttl(self, request_data):
        if self._get_lexicon_option("ttl"):
            request_data["ttl"] = self._get_lexicon_option("ttl")

    # From Glesys record structure: [u'domainname', u'recordid', u'type', u'host', u'ttl', u'data']
    def _glesysrecord2lexiconrecord(self, glesys_record):
        return {
            "id": glesys_record["recordid"],
            "type": glesys_record["type"],
            "name": glesys_record["host"],
            "ttl": glesys_record["ttl"],
            "content": glesys_record["data"],
        }



"""Module provider for Aurora"""
import base64
import datetime
import hashlib
import hmac
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["auroradns.eu"]


def provider_parser(subparser):
    """Configure provider parser for Aurora"""
    subparser.add_argument("--auth-api-key", help="specify API key for authentication")
    subparser.add_argument(
        "--auth-secret-key", help="specify the secret key for authentication"
    )


class Provider(BaseProvider):
    """Provider for Aurora"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.auroradns.eu"

    def _authenticate(self):
        payload = self._get("/zones")

        for item in payload:
            if item["name"] == self.domain:
                self.domain_id = item["id"]
                break
        else:
            raise AuthenticationError("No domain found")

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        data = {"type": rtype, "name": self._relative_name(name), "content": content}
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")
        payload = self._post(f"/zones/{self.domain_id}/records", data)

        LOGGER.debug("create_record: %s", payload)
        return payload

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get(f"/zones/{self.domain_id}/records")

        # Apply filtering first.
        processed_records = payload
        if rtype:
            processed_records = [
                record for record in processed_records if record["type"] == rtype
            ]
        if name:
            processed_records = [
                record
                for record in processed_records
                if record["name"] == self._relative_name(name)
            ]
        if content:
            processed_records = [
                record
                for record in processed_records
                if record["content"].lower() == content.lower()
            ]

        # Format the records.
        records = []
        for record in processed_records:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["content"],
                "id": record["id"],
            }
            records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        # Try to find record if no identifier was specified
        if not identifier:
            identifier = self._find_record_identifier(rtype, name, None)

        data = {}
        if rtype:
            data["type"] = rtype
        if name:
            data["name"] = self._relative_name(name)
        if content:
            data["content"] = content
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        payload = self._put(f"/zones/{self.domain_id}/records/{identifier}", data)

        LOGGER.debug("update_record: %s", payload)
        return payload

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        # Try to find record if no identifier was specified
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/zones/{self.domain_id}/records/{record_id}")

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        time = datetime.datetime.utcnow()
        timestamp = time.strftime("%Y%m%dT%H%M%SZ")
        authorization_header = self._generate_auth_header(action, url, timestamp)

        request = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers={
                "X-AuroraDNS-Date": timestamp,
                "Authorization": authorization_header,
                "Content-Type": "application/json",
            },
        )

        # If the response is a HTTP 409 statusCode, the record already exists: return true.
        if request.status_code == 409:
            return True

        # If the request fails for any other reason, throw an error.
        request.raise_for_status()

        # Try to parse the json, if it not exists, return true.
        try:
            return request.json()
        except BaseException:
            return True

    def _generate_auth_header(self, action, url, timestamp):
        secret_key = self._get_provider_option("auth_secret_key")
        api_key = self._get_provider_option("auth_api_key")
        sig = action + url + timestamp

        signature = base64.b64encode(
            hmac.new(
                secret_key.encode("utf-8"),
                sig.encode("utf-8"),
                digestmod=hashlib.sha256,
            ).digest()
        )

        auth = api_key + ":" + signature.decode("utf-8")
        auth_b64 = base64.b64encode(auth.encode("utf-8"))
        return f"AuroraDNSv1 {auth_b64.decode('utf-8')}"

    def _find_record_identifier(self, rtype, name, content):
        records = self._list_records(rtype, name, content)
        LOGGER.debug("records: %s", records)
        if len(records) == 1:
            return records[0]["id"]
        raise Exception(
            "Record identifier could not be found. Try to provide an identifier"
        )



"""
This module takes care of finding information about the runtime of Lexicon:
* what are the providers installed, and available
* what is the version of Lexicon
"""
import pkgutil
import re
from typing import Dict

try:
    from importlib.metadata import Distribution, PackageNotFoundError
except ModuleNotFoundError:
    from importlib_metadata import Distribution, PackageNotFoundError  # type: ignore[misc]

from lexicon import providers


def find_providers() -> Dict[str, bool]:
    """Find all providers registered in Lexicon, and their availability"""
    providers_list = sorted(
        {
            modname
            for (_, modname, _) in pkgutil.iter_modules(providers.__path__)  # type: ignore
            if modname != "base"
        }
    )

    try:
        distribution = Distribution.from_name("dns-lexicon")
    except PackageNotFoundError:
        return {provider: True for provider in providers_list}
    else:
        return {
            provider: _resolve_requirements(provider, distribution)
            for provider in providers_list
        }


def lexicon_version() -> str:
    """Retrieve current Lexicon version"""
    try:
        return Distribution.from_name("dns-lexicon").version
    except PackageNotFoundError:
        return "unknown"


def _resolve_requirements(provider: str, distribution: Distribution) -> bool:
    requires = distribution.requires
    if requires is None:
        raise ValueError("Error while trying finding requirements.")

    requirements = [
        re.sub(r"^(.*)\s\(.*\)(?:;.*|)$", r"\1", requirement)
        for requirement in requires
        if f'extra == "{provider}"' in requirement
    ]

    if not requirements:
        # No extra for this provider
        return True

    for requirement in requirements:
        try:
            Distribution.from_name(requirement)
        except PackageNotFoundError:
            # At least one extra requirement is not fulfilled
            return False

    # All extra requirements are fulfilled
    return True



"""Module provider for Webgo"""
import logging

from bs4 import BeautifulSoup  # type: ignore
from requests import Session

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["webgo.de"]


def provider_parser(subparser):
    """Configure provider parser for Webgo."""
    subparser.description = """A provider for Webgo."""
    subparser.add_argument(
        "--auth-username", help="specify username for authentication"
    )
    subparser.add_argument(
        "--auth-password", help="specify password for authentication"
    )


class Provider(BaseProvider):
    """
    webgo.de provider
    """

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain = self.domain
        self.domain_id = None
        self.session = None

    def _authenticate(self):
        # Create the session GET the login page to retrieve a session cookie
        self.session = Session()
        self.session.get("https://login.webgo.de/")

        # Hit the login page with authentication info to login the session
        login_response = self.session.post(
            "https://login.webgo.de/login",
            data={
                "data[User][username]": self._get_provider_option("auth_username")
                or "",
                "data[User][password]": self._get_provider_option("auth_password")
                or "",
            },
        )

        # Parse in the HTML, if the div containing the error message is found, error
        html = BeautifulSoup(login_response.content, "html.parser")
        if html.find("div", {"class": "loginformerror"}) is not None:
            LOGGER.warning("Webgo login failed, check Username and Password")
            raise AuthenticationError("Webgo login failed, check Username and Password")
            return False

        # Make an authenticated GET to the DNS management page
        zones_response = self.session.get("https://login.webgo.de/domains")

        html = BeautifulSoup(zones_response.content, "html.parser")
        domain_table = html.find("table", {"class": "alltable"})
        rows = domain_table.find_all("tr")
        dns_link = None
        for row in rows[1:]:
            domain = row.findAll("td")[1].renderContents().decode()
            if domain == self.domain:
                dns_link = row.findAll("td")[5]
                dns_link = dns_link.find(
                    "a", {"class": "domainButton fcon-sliders"}
                ).get("href")

        # If the Domain couldn't be found, error, otherwise, return the value of the tag
        if dns_link is None:
            LOGGER.warning("Domain %s not found in account", self.domain)
            raise AuthenticationError(f"Domain {self.domain} not found in account")

        self.domain_id = dns_link.rsplit("/", 1)[1]
        LOGGER.debug("Webgo domain ID: %s", self.domain_id)
        return True

    # Create record. If record already exists with the same content, do nothing
    def _create_record(self, rtype, name, content):
        LOGGER.debug("Creating record for zone %s", name)
        # Pull a list of records and check for ours
        if name:
            if name == self.domain:
                LOGGER.warning(
                    "Unable to create record because your main domain %s can't be re-created",
                    self.domain,
                )
                return False
            name = self._relative_name(name)
        if rtype == "CNAME" and not content.endswith("."):
            content += "."
        records = self._list_records(rtype=rtype, name=name, content=content)
        if len(records) >= 1:
            LOGGER.warning("Duplicate record %s %s %s, NOOP", rtype, name, content)
            return True
        data = {
            "data[DnsSetting][sub]": name,
            "data[DnsSetting][ttl]": "3600",
            "data[DnsSetting][rr-typ]": rtype,
            "data[DnsSetting][pref-mx]": "0",
            "data[DnsSetting][value]": content,
            "data[DnsSetting][action]": "newsub",
            "data[DnsSetting][domain_id]": self.domain_id,
        }
        ttl = self._get_lexicon_option("ttl")
        if ttl:
            if ttl <= 0:
                data["data[DnsSetting][ttl]"] = "3600"
            else:
                data["data[DnsSetting][ttl]"] = str(ttl)
        prio = self._get_lexicon_option("priority")
        if prio:
            if prio <= 0:
                data["data[DnsSetting][pref-mx]"] = "10"
            else:
                data["data[DnsSetting][pref-mx]"] = str(prio)

        self.session.post(
            "https://login.webgo.de/dns_settings/domainDnsEditForm", data=data
        )
        self.session.get(
            f"https://login.webgo.de/dnsSettings/domainDnsDo/{self.domain_id}/ok"
        )
        # Pull a list of records and check for ours
        records = self._list_records(name=name)
        if len(records) >= 1:
            LOGGER.info("Successfully added record %s", name)
            return True
        LOGGER.info("Failed to add record %s", name)
        return False

    # List all records. Return an empty list if no records found.
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is
    # received.
    def _list_records(self, rtype=None, name=None, content=None):
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(
        self, rtype=None, name=None, content=None, identifier=None
    ):
        if name:
            name = self._relative_name(name)
        records = []
        # Make an authenticated GET to the DNS management page
        edit_response = self.session.get(
            f"https://login.webgo.de/dnsSettings/domainDnsEdit/{self.domain_id}",
        )

        # Parse the HTML response, and list the table rows for DNS records
        html = BeautifulSoup(edit_response.content, "html.parser")
        dns_table = html.find("table", {"class": "alltable"})
        records = dns_table.findAll("tr")
        # If the tag couldn't be found, error, otherwise, return the value of the tag
        if records is None or not records:
            LOGGER.warning("Domains not found in account")
            return records
        new_records = []
        # Find Main Record
        rec = {}
        mainip = html.find("span", {"class": "mainIp"})
        mainip_record = mainip.find_next("span").text
        dns_link = mainip.find_next("a").get("href")
        rec["name"] = self.domain
        rec["ttl"] = "3600"
        rec["type"] = "A"
        rec["prio"] = "10"
        rec["content"] = mainip_record
        rec["id"] = dns_link.rsplit("/", 2)[1]
        rec["option"] = "main"
        new_records.append(rec)
        # Find Subrecords
        for dns_tr in records[1:]:
            tds = dns_tr.findAll("td")
            # Process HTML in the TR children to derive each object
            rec = {}
            rec["name"] = self._full_name(tds[0].string)
            rec["ttl"] = tds[1].string
            rec["type"] = tds[2].string
            rec["prio"] = tds[3].string
            rec["content"] = tds[4].string
            dns_link = tds[5]
            dns_link = dns_link.find("a", {"class": "domainButton fcon-edit"}).get(
                "href"
            )
            rec["id"] = dns_link.rsplit("/", 2)[1]
            if rec["content"].startswith('"'):
                rec = self._clean_TXT_record(rec)
            rec["option"] = "sub"
            new_records.append(rec)
        records = new_records
        if identifier:
            LOGGER.debug("Filtering %d records by id: %s", len(records), identifier)
            records = [
                record for record in records if str(record["id"]) == str(identifier)
            ]
        if rtype:
            LOGGER.debug("Filtering %d records by rtype: %s", len(records), rtype)
            records = [record for record in records if record["type"] == rtype]
        if name:
            LOGGER.debug("Filtering %d records by name: %s", len(records), name)
            if name.endswith("."):
                name = name[:-1]
            records = [record for record in records if name in record["name"]]
        if content:
            LOGGER.debug(
                "Filtering %d records by content: %s", len(records), content.lower()
            )
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]
        LOGGER.debug("Final records (%d): %s", len(records), records)

        return records

    # Create or update a record.
    def _update_record(self, identifier=None, rtype=None, name=None, content=None):
        maindata = None
        sub_update = None
        if identifier is not None:
            records = self._list_records_internal(identifier=identifier)
        else:
            records = self._list_records_internal(name=name, rtype=rtype)
        for record in records:
            # Check whether Main-Domain needs to be updated
            if record["option"] == "main":
                maindata = {
                    "data[DnsSetting][value]": content,
                    "data[DnsSetting][action]": "main",
                    "data[DnsSetting][domain_id]": record["id"],
                }
            # Update every Subrecord
            else:
                # Delete record if it exists
                # Record ID is changed after Update from main!
                self._delete_record(identifier=record["id"])
                self._create_record(record["type"], record["name"], content)
                sub_update = True
        # Check whether we need to update main
        if maindata is not None:
            # If we updated an Subdomain in the meantime, the ID changed and we need to refresh it for the main-domain
            if sub_update is not None:
                if identifier is not None:
                    records = self._list_records_internal(identifier=identifier)
                else:
                    records = self._list_records_internal(name=name, rtype=rtype)
                for record in records:
                    # Check whether Main-Domain needs to be updated
                    if record["option"] == "main":
                        maindata = {
                            "data[DnsSetting][value]": content,
                            "data[DnsSetting][action]": "main",
                            "data[DnsSetting][domain_id]": record["id"],
                        }
                self.session.post(
                    "https://login.webgo.de/dns_settings/domainDnsEditForm",
                    data=maindata,
                )
                self.session.get(
                    f"https://login.webgo.de/dnsSettings/domainDnsDo/{self.domain_id}/ok"
                )
                LOGGER.debug("Updated Main Domain %s", records[0]["name"])
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_ids = []
        records = self._list_records_internal(rtype, name, content, identifier)
        if "main" in [record["option"] for record in records]:
            LOGGER.warning(
                "Unable to delete records because your main domain %s can't be deleted",
                self.domain,
            )
            return False
        delete_record_ids = [record["id"] for record in records]
        LOGGER.debug("Record IDs to delete: %s", delete_record_ids)
        for rec_id in delete_record_ids:
            response = self.session.get(
                f"https://login.webgo.de/dnsSettings/domainDnsDo/{rec_id}/delete"
            )
            if response.status_code == 200:
                self.session.get(
                    f"https://login.webgo.de/dnsSettings/domainDnsDo/{self.domain_id}/ok"
                )
            else:
                LOGGER.warning("Unable to delete record %s", rec_id)
                return False
        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used in this provider
        pass



"""Module provider for Online.net"""
import json
import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["online.net"]


def provider_parser(subparser):
    """Configure provider parser for Online.net"""
    subparser.add_argument("--auth-token", help="specify private api token")


def _to_data(rtype, content):
    if rtype == "TXT":
        return f'"{content}"'
    return content


class Provider(BaseProvider):
    """Provider class for Online.net"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.zone_name = "Zone Automatic Lexicon "
        self.passive_zone = None
        self.active_zone = None
        self.domain_id = self.domain
        self.api_endpoint = "https://api.online.net/api/v1"

    def _authenticate(self):
        self._init_zones()

    def _list_zones(self):
        return self._get(f"/domain/{self.domain_id}/version")

    def _init_zones(self):
        # sets current zone version
        zone_name_a = self.zone_name + "A"
        zone_name_b = self.zone_name + "B"
        active_row = None
        passive_row = None
        for row in self._list_zones():
            if row["active"]:
                active_row = row
            elif row["name"] == zone_name_a or row["name"] == zone_name_b:
                passive_row = row

        if passive_row is None:
            passive_row = self._post(
                f"/domain/{self.domain_id}/version",
                {
                    "name": zone_name_b
                    if active_row["name"] == zone_name_a
                    else zone_name_a
                },
            )

        self.active_zone = active_row["uuid_ref"]
        self.passive_zone = passive_row["uuid_ref"]
        self._update_passive_zone()

    def _update_passive_zone(self):
        self._put(
            f"/domain/{self.domain_id}/version/{self.passive_zone}/zone_from_bind",
            self._get_bind_zone(),
        )

    def _get_bind_zone(self):
        records = self._list_zone_records(self.active_zone)
        # then convert records to bind format
        bind_str = ""
        for record in records:
            bind_str = bind_str + "{0} {1} IN {2} {3}{4}\n".format(
                record["name"] or "@",
                record["ttl"],
                record["type"],
                f"{record['aux']} " if "aux" in record else "",
                record["data"] or "",
            )
        return bind_str

    def _enable_zone(self):
        zone = self.passive_zone
        if zone is None:
            raise Exception("Could not enable uninitialized passive_zone")
        self._patch(f"/domain/{self.domain_id}/version/{zone}/enable")
        self.passive_zone = self.active_zone
        self.active_zone = zone
        self._update_passive_zone()

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        try:
            record = self._find_record(rtype, name, content)
            if record is not None:
                return True

            record = {
                "name": self._fqdn_name(name),
                "type": rtype,
                "data": _to_data(rtype, content),
                "priority": self._get_lexicon_option("priority") or "",
                "ttl": self._get_lexicon_option("ttl") or "",
            }

            self._post(
                f"/domain/{self.domain_id}/version/{self.passive_zone}/zone", record
            )
        except BaseException as error:
            LOGGER.debug(error)
            return False

        self._enable_zone()
        LOGGER.debug("create_record: %s", True)
        return True

    def _find_zone_records(self, zone, rtype=None, name=None, content=None):
        records = []
        for record in self._list_zone_records(zone):
            processed_record = {
                "id": record["id"],
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["data"],
                "priority": record["aux"] if "aux" in record else "",
            }
            records.append(self._clean_TXT_record(processed_record))

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            full_name = self._full_name(name)
            records = [record for record in records if record["name"] == full_name]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    def _list_zone_records(self, zone_id):
        return self._get(f"/domain/{self.domain_id}/version/{zone_id}/zone")

    def _list_records(self, rtype=None, name=None, content=None):
        return self._find_zone_records(self.passive_zone, rtype, name, content)

    def _find_record(self, rtype=None, name=None, content=None):
        records = self._list_records(rtype, name, content)
        if not records:
            return None
        return records[0]

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        record = self._find_record(rtype, name)
        if record is None:
            LOGGER.debug(
                "cannot find record to update: %s %s %s", identifier, rtype, name
            )
            return True
        if rtype:
            record["type"] = rtype
        if name:
            record["name"] = self._fqdn_name(name)
        if content:
            record["data"] = _to_data(rtype, content)
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")
        # it is weird that 'aux' becomes 'priority' in online's api
        if self._get_lexicon_option("priority"):
            record["priority"] = self._get_lexicon_option("priority")

        if identifier is None:
            identifier = record["id"]

        record.pop("id")

        try:
            self._patch(
                f"/domain/{self.domain_id}/version/{self.passive_zone}/zone/{identifier}",
                record,
            )

        except BaseException as error:
            LOGGER.debug(error)
            return False

        self._enable_zone()
        # If it didn't raise from the http status code, then we're good
        LOGGER.debug("update_record: %s", identifier)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        records = self._list_records(rtype, name, content)
        if not records:
            LOGGER.debug("Cannot find records %s %s %s", rtype, name, content)
            return False
        LOGGER.debug("delete_records: %s records found", len(records))
        try:
            for record in records:
                self._delete(
                    f"/domain/{self.domain_id}/version/{self.passive_zone}/zone/{record['id']}"
                )
        except BaseException as error:
            LOGGER.debug(error)
            return False

        self._enable_zone()
        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    def _patch(self, url="/", data=None, query_params=None):
        return self._request("PATCH", url, data=data, query_params=query_params)

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if query_params is None:
            query_params = {}

        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self._get_provider_option('auth_token')}",
        }
        if data is not None:
            if isinstance(data, str):
                headers["Content-Type"] = "text/plain"
            else:
                headers["Content-Type"] = "application/json"
                data = json.dumps(data)

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=data,
            headers=headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()

        return response.json() if response.text else ""



"""Module provider for nfsn"""
import hashlib
import logging
import random
import string
import time
from urllib.parse import urlencode

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["nearlyfreespeech.net"]


def provider_parser(subparser):
    """Generate subparser for nfsn"""
    subparser.add_argument(
        "--auth-username", help="specify username used to authenticate"
    )
    subparser.add_argument("--auth-token", help="specify token used to authenticate")


SALT_SHAKER = string.ascii_letters + string.digits


class Provider(BaseProvider):
    """Provider class for nfsn"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.nearlyfreespeech.net"
        self.shortname = None

    def _authenticate(self):
        try:
            self._post(f"/dns/{self.domain}/listRRs")
        except Exception as e:
            raise AuthenticationError(e)
        self.domain_id = self.domain

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        existing_records = self.list_records(rtype, name, content)
        if existing_records:
            # Do nothing if the record already exists.
            # The creation call can fail for a variety of reasons, so
            # the safest thing to do is check if the record already exists
            return True

        self._do_create(rtype, name, content)
        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        params = {}
        if rtype is not None:
            params["type"] = rtype
        if name is not None:
            params["name"] = self._relative_name(name)
        if content is not None:
            params["data"] = content

        url = f"/dns/{self.domain_id}/listRRs"
        records = self._post(url, params)
        records = [
            {
                "type": r["type"],
                "name": self._full_name(r["name"]),
                "ttl": r["ttl"],
                "content": r["data"],
                "id": hashlib.sha1(
                    "".join([r["type"], r["name"], r["data"]]).encode("utf-8")
                ).hexdigest(),
            }
            for r in records
        ]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is not None:
            records = self._list_records()
            to_delete = next((r for r in records if r["id"] == identifier), None)
            if to_delete is None:
                raise ValueError("No record with that identifier.")
        else:
            # Check name and rtype
            matching_records = self._list_records(rtype=rtype, name=name)
            if len(matching_records) > 1:
                raise ValueError(
                    "More than one record exists with that type and name. "
                    "Try specifying an identifier."
                )
            to_delete = matching_records[0]

        self._do_delete(to_delete["type"], to_delete["name"], to_delete["content"])
        self._do_create(rtype, name, content)
        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        matching_records = self._list_records(rtype, name, content)
        if identifier is not None:
            to_delete = next(
                (r for r in matching_records if r["id"] == identifier), None
            )
            if to_delete is None:
                raise ValueError("No record with that identifier.")
            to_delete = [to_delete]
        else:
            to_delete = matching_records

        for record in to_delete:
            self._do_delete(record["type"], record["name"], record["content"])

        LOGGER.debug("delete_record: %s", True)
        return True

    def _do_create(self, rtype, name, content):
        record = {"name": self._relative_name(name), "type": rtype, "data": content}

        ttl = self._get_lexicon_option("ttl")
        if ttl:
            record["ttl"] = ttl

        self._post(f"/dns/{self.domain_id}/addRR", record)
        return True

    def _do_delete(self, rtype=None, name=None, content=None):
        url = f"/dns/{self.domain_id}/removeRR"
        record = {"name": self._relative_name(name), "type": rtype, "data": content}
        self._post(url, record)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is not None:
            body = urlencode(data)
            hashed_body = hashlib.sha1(body.encode("utf-8")).hexdigest()
        else:
            hashed_body = hashlib.sha1().hexdigest()

        timestamp = str(int(time.time()))
        salt = "".join(random.choice(SALT_SHAKER) for _ in range(16))
        hash_items = [
            self._get_provider_option("auth_username"),
            timestamp,
            salt,
            self._get_provider_option("auth_token"),
            url,
            hashed_body,
        ]
        auth_hash = hashlib.sha1(";".join(hash_items).encode("utf-8")).hexdigest()
        auth_value = ";".join(
            [self._get_provider_option("auth_username"), timestamp, salt, auth_hash]
        )
        auth_header = {"X-NFSN-Authentication": auth_value}

        response = requests.request(
            action, "".join([self.api_endpoint, url]), data=data, headers=auth_header
        )
        response.raise_for_status()
        if response.content:
            return response.json()
        return {}



"""Module provider for OVH"""
import hashlib
import json
import logging
import time

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

ENDPOINTS = {
    "ovh-eu": "https://eu.api.ovh.com/1.0",
    "ovh-ca": "https://ca.api.ovh.com/1.0",
    "ovh-us": "https://api.ovhcloud.com/1.0",
    "kimsufi-eu": "https://eu.api.kimsufi.com/1.0",
    "kimsufi-ca": "https://ca.api.kimsufi.com/1.0",
    "soyoustart-eu": "https://eu.api.soyoustart.com/1.0",
    "soyoustart-ca": "https://ca.api.soyoustart.com/1.0",
}

NAMESERVER_DOMAINS = ["ovh.net", "anycast.me"]


def provider_parser(subparser):
    """Generate provider parser for OVH"""
    subparser.description = """
        OVH Provider requires a token with full rights on /domain/*.
        It can be generated for your OVH account on the following URL:
        https://api.ovh.com/createToken/index.cgi?GET=/domain/*&PUT=/domain/*&POST=/domain/*&DELETE=/domain/*"""
    subparser.add_argument(
        "--auth-entrypoint",
        help="specify the OVH entrypoint",
        choices=[
            "ovh-eu",
            "ovh-ca",
            "soyoustart-eu",
            "soyoustart-ca",
            "kimsufi-eu",
            "kimsufi-ca",
        ],
    )
    subparser.add_argument("--auth-application-key", help="specify the application key")
    subparser.add_argument(
        "--auth-application-secret", help="specify the application secret"
    )
    subparser.add_argument("--auth-consumer-key", help="specify the consumer key")


class Provider(BaseProvider):
    """Provider class for OVH"""

    def __init__(self, config):
        super(Provider, self).__init__(config)

        # Handling missing required parameters
        if not self._get_provider_option("auth_entrypoint"):
            raise Exception("Error, entrypoint is not defined")
        if not self._get_provider_option("auth_application_key"):
            raise Exception("Error, application key is not defined")
        if not self._get_provider_option("auth_application_secret"):
            raise Exception("Error, application secret is not defined")
        if not self._get_provider_option("auth_consumer_key"):
            raise Exception("Error, consumer key is not defined")

        # Construct DNS OVH environment
        self.domain_id = None
        self.endpoint_api = ENDPOINTS.get(self._get_provider_option("auth_entrypoint"))
        self.session = None
        self.time_delta = None

    def _authenticate(self):
        # All requests will be done in one HTTPS session
        self.session = requests.Session()

        # Calculate delta time between local and OVH to avoid requests rejection
        server_time = self.session.get(f"{self.endpoint_api}/auth/time").json()
        self.time_delta = server_time - int(time.time())

        # Get domain and status
        domain = self.domain

        domains = self._get("/domain/zone/")
        if domain not in domains:
            raise AuthenticationError(f"Domain {domain} not found")

        status = self._get(f"/domain/zone/{domain}/status")
        if not status["isDeployed"]:
            raise AuthenticationError(f"Zone {domain} is not deployed")

        self.domain_id = domain

    def _create_record(self, rtype, name, content):
        domain = self.domain
        ttl = self._get_lexicon_option("ttl")

        records = self._list_records(rtype, name, content)
        for record in records:
            if (
                record["type"] == rtype
                and self._relative_name(record["name"]) == self._relative_name(name)
                and record["content"] == content
            ):
                LOGGER.debug(
                    "create_record (ignored, duplicate): %s %s %s", rtype, name, content
                )
                return True

        data = {
            "fieldType": rtype,
            "subDomain": self._relative_name(name),
            "target": content,
        }

        if ttl:
            data["ttl"] = ttl

        result = self._post(f"/domain/zone/{domain}/record", data)
        self._post(f"/domain/zone/{domain}/refresh")

        LOGGER.debug("create_record: %s", result["id"])

        return True

    def _list_records(self, rtype=None, name=None, content=None):
        domain = self.domain
        records = []

        params = {}
        if rtype:
            params["fieldType"] = rtype
        if name:
            params["subDomain"] = self._relative_name(name)

        record_ids = self._get(f"/domain/zone/{domain}/record", params)

        for record_id in record_ids:
            raw = self._get(f"/domain/zone/{domain}/record/{record_id}")
            records.append(
                {
                    "type": raw["fieldType"],
                    "name": self._full_name(raw["subDomain"]),
                    "ttl": raw["ttl"],
                    "content": raw["target"],
                    "id": raw["id"],
                }
            )

        if content:
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", records)

        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        domain = self.domain

        if not identifier:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]
            elif len(records) > 1:
                raise Exception("Several record identifiers match the request")
            else:
                raise Exception("Record identifier could not be found")

        data = {}
        if name:
            data["subDomain"] = self._relative_name(name)
        if content:
            data["target"] = content

        self._put(f"/domain/zone/{domain}/record/{identifier}", data)
        self._post(f"/domain/zone/{domain}/refresh")

        LOGGER.debug("update_record: %s", identifier)

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        domain = self.domain

        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/domain/zone/{domain}/record/{record_id}")

        self._post(f"/domain/zone/{domain}/refresh")

        LOGGER.debug("delete_record: %s", True)

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        headers = {}
        target = self.endpoint_api + url
        body = ""

        if data is not None:
            headers["Content-type"] = "application/json"
            body = json.dumps(data)

        # Get correctly sync time
        now = str(int(time.time()) + self.time_delta)

        headers["X-Ovh-Application"] = self._get_provider_option("auth_application_key")
        headers["X-Ovh-Consumer"] = self._get_provider_option("auth_consumer_key")
        headers["X-Ovh-Timestamp"] = now

        request = requests.Request(
            action, target, data=body, params=query_params, headers=headers
        )
        prepared_request = self.session.prepare_request(request)

        # Build OVH API signature for the current request
        signature = hashlib.sha1()
        signature.update(
            "+".join(
                [
                    self._get_provider_option("auth_application_secret"),
                    self._get_provider_option("auth_consumer_key"),
                    action.upper(),
                    prepared_request.url,
                    body,
                    now,
                ]
            ).encode("utf-8")
        )

        # Sign the request
        headers["X-Ovh-Signature"] = "$1$" + signature.hexdigest()

        result = self.session.request(
            method=action, url=target, params=query_params, data=body, headers=headers
        )
        result.raise_for_status()

        return result.json()



"""Module provider for CloudXNS"""
import hashlib
import json
import logging
import time
from urllib.parse import urlencode

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["cloudxns.net"]


def provider_parser(subparser):
    """Generate subparser for CloudXNS"""
    subparser.add_argument("--auth-username", help="specify API-KEY for authentication")
    subparser.add_argument("--auth-token", help="specify SECRET-KEY for authentication")


class Provider(BaseProvider):
    """Provider class for CloudXNS"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://www.cloudxns.net/api2"

    def _authenticate(self):

        payload = self._get("/domain")
        for record in payload["data"]:
            if record["domain"] == self.domain + ".":
                self.domain_id = record["id"]
                break
        else:
            raise AuthenticationError("No domain found")

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):

        record = {
            "domain_id": self.domain_id,
            "host": self._relative_name(name),
            "value": content,
            "type": rtype,
            "line_id": 1,
        }
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        try:
            self._post("/record", record)
        except requests.exceptions.HTTPError as err:
            already_exists = err.response.json()["code"] == 34
            if not already_exists:
                raise

        # CloudXNS will return bad HTTP Status when error, will throw at
        # r.raise_for_status() in _request()
        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get(
            "/record/" + self.domain_id, {"host_id": 0, "offset": 0, "row_num": 2000}
        )
        records = []
        for record in payload["data"]:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["host"]),
                "ttl": record["ttl"],
                "content": record["value"],
                # this id is useless unless your doing record linking. Lets return the
                # original record identifier.
                "id": record["record_id"],
            }
            if processed_record["type"] == "TXT":
                processed_record["content"] = processed_record["content"].replace(
                    '"', ""
                )
                # CloudXNS will add quotes automaticly for TXT records,
                # https://www.cloudxns.net/Support/detail/id/114.html
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        if not identifier:
            records = self._list_records(name=name)
            if len(records) == 1:
                identifier = records[0]["id"]
            else:
                raise Exception("Record identifier could not be found.")

        data = {
            "domain_id": self.domain_id,
            "host": self._relative_name(name),
            "value": content,
            "type": rtype,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        self._put("/record/" + identifier, data)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete("/record/" + record_id + "/" + self.domain_id)

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        data["login_token"] = (
            self._get_provider_option("auth_username")
            + ","
            + self._get_provider_option("auth_token")
        )
        data["format"] = "json"
        if query_params:
            query_string = "?" + urlencode(query_params)
        else:
            query_string = ""
            query_params = {}
        if data:
            data = json.dumps(data)
        else:
            data = ""
        date = time.strftime("%a %b %d %H:%M:%S %Y", time.localtime())
        default_headers = {
            "API-KEY": self._get_provider_option("auth_username"),
            "API-REQUEST-DATE": date,
            "API-HMAC": hashlib.md5(
                f"{self._get_provider_option('auth_username')}{self.api_endpoint}{url}{query_string}{data}{date}{self._get_provider_option('auth_token')}".encode(
                    "utf-8"
                )
            ).hexdigest(),
            "API-FORMAT": "json",
        }
        default_auth = None
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=data,
            headers=default_headers,
            auth=default_auth,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Module provider for Njalla"""
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["1-you.njalla.no", "2-can.njalla.in", "3-get.njalla.fo"]


def provider_parser(subparser):
    """Module provider for Njalla"""
    subparser.add_argument("--auth-token", help="specify API token for authentication")


class Provider(BaseProvider):
    """Provider class for Njalla"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://njal.la/api/1/"

    def _authenticate(self):
        params = {"domain": self.domain}
        try:
            result = self._api_call("get-domain", params)
        except Exception as e:
            raise AuthenticationError(str(e))

        if result["name"] != self.domain:
            raise AuthenticationError("Domain not found")

        self.domain_id = self.domain

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        params = {
            "domain": self.domain,
            "type": rtype,
            "name": name,
            "content": content,
            "ttl": 10800,
        }
        if self._get_lexicon_option("ttl"):
            params["ttl"] = self._get_lexicon_option("ttl")
        result = self._api_call("add-record", params)

        LOGGER.debug("create_record: %s", result)
        return result

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        params = {"domain": self.domain}
        result = self._api_call("list-records", params)

        records = result["records"]
        processed_records = [
            {
                "id": record["id"],
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["content"],
            }
            for record in records
        ]
        filtered_records = [
            record
            for record in processed_records
            if (
                (rtype is None or record["type"] == rtype)
                and (name is None or record["name"] == self._full_name(name))
                and (content is None or record["content"] == content)
            )
        ]

        LOGGER.debug("list_records: %s", filtered_records)
        return filtered_records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            identifier = self._get_record_identifier(rtype=rtype, name=name)

        params = {"id": identifier, "domain": self.domain, "content": content}
        result = self._api_call("edit-record", params)

        LOGGER.debug("update_record: %s", result)
        return result

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            identifier = self._get_record_identifier(
                rtype=rtype, name=name, content=content
            )

        params = {"domain": self.domain, "id": identifier}
        self._api_call("remove-record", params)

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _api_call(self, method, params):
        if self._get_provider_option("auth_token") is None:
            raise Exception("Must provide API token")

        data = {"method": method, "params": params}
        response = self._request("POST", "", data)

        if "error" in response.keys():
            error = response["error"]
            raise Exception("%d: %s" % (error["code"], error["message"]))

        return response["result"]

    def _get_record_identifier(self, rtype=None, name=None, content=None):
        records = self._list_records(rtype=rtype, name=name, content=content)
        if len(records) == 1:
            return records[0]["id"]

        raise Exception("Unambiguous record could not be found.")

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        token = self._get_provider_option("auth_token")
        headers = {
            "Authorization": "Njalla " + token,
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        response = requests.request(
            action,
            self.api_endpoint + url,
            headers=headers,
            params=query_params,
            json=data,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["porkbun.com"]


def provider_parser(subparser):
    """Return the parser for this provider"""
    subparser.description = """
        To authenticate with Porkbun, you need both an API key and a
        secret API key. These can be created at porkbun.com/account/api .
    """

    subparser.add_argument("--auth-key", help="specify API key for authentication")
    subparser.add_argument(
        "--auth-secret", help="specify secret API key for authentication"
    )


class Provider(BaseProvider):
    """Provider class for Porkbun"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.api_endpoint = "https://porkbun.com/api/json/v3"
        self._api_key = self._get_provider_option("auth_key")
        self._secret_api_key = self._get_provider_option("auth_secret")
        self._auth_data = {
            "apikey": self._api_key,
            "secretapikey": self._secret_api_key,
        }

        self.domain = self._get_lexicon_option("domain")

    def _authenticate(self):
        # more of a test that the authentication works
        response = self._post("/ping")

        if response["status"] != "SUCCESS":
            raise AuthenticationError("Incorrect API keys")
        self.domain_id = self.domain
        self._list_records()

    def _create_record(self, rtype, name, content):
        active_records = self._list_records(rtype, name, content)
        # if the record already exists: early exit, return success
        if active_records:
            LOGGER.debug("create_record: record already exists")
            return True

        data = {
            "type": rtype,
            "content": content,
            "name": self._relative_name(name),
        }

        if self._get_lexicon_option("priority"):
            data["prio"] = self._get_lexicon_option("priority")

        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        response = self._post(f"/dns/create/{self.domain}", data)

        LOGGER.debug(f"create_record: {response}")
        return response["status"] == "SUCCESS"

    def _list_records(self, rtype=None, name=None, content=None):
        # porkbun has some weird behavior on the retrieveByNameType endpoint
        # related to how it handles subdomains.
        # so we ignore it and filter locally instead
        records = self._post(f"/dns/retrieve/{self.domain}")

        if records["status"] != "SUCCESS":
            raise requests.exceptions.HTTPError(records)

        records = records["records"]

        records = self._format_records(records)

        # filter for content if it was provided
        if content is not None:
            records = [x for x in records if x["content"] == content]

        # filter for name if it was provided
        if name is not None:
            records = [x for x in records if x["name"] == self._full_name(name)]

        # filter for rtype if it was provided
        if rtype is not None:
            records = [x for x in records if x["type"] == rtype]

        LOGGER.debug(f"list_records: {records}")
        LOGGER.debug(f"Number of records retrieved: {len(records)}")
        return records

    def _update_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier is None:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]
            elif len(records) == 0:
                raise Exception(
                    "No records found matching type and name - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type and name - won't update"
                )

        endpoint = f"/dns/edit/{self.domain}/{identifier}"

        data = {"name": self._relative_name(name), "type": rtype, "content": content}

        # if set to 0, then this will automatically get set to 300
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        if self._get_lexicon_option("priority"):
            data["prio"] = self._get_lexicon_option("priority")

        result = self._post(endpoint, data)

        LOGGER.debug(f"update_record: {result}")
        return result["status"] == "SUCCESS"

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier is None:
            records = self._list_records(rtype, name, content)
            delete_record_ids = [record["id"] for record in records]
        else:
            delete_record_ids = [identifier]

        LOGGER.debug(f"deleting records: {delete_record_ids}")

        for record_id in delete_record_ids:
            self._post(f"/dns/delete/{self.domain}/{record_id}")

        LOGGER.debug("delete_record: success")
        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        headers = {"Content-Type": "application/json"}

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps({**data, **self._auth_data}),
            headers=headers,
        )

        response.raise_for_status()
        return response.json()

    def _format_records(self, records):
        for record in records:
            record["name"] = self._full_name(record["name"])
            if "prio" in record:
                record["options"] = {"mx": {"priority": record["prio"]}}
                del record["prio"]
        return records



"""Module provider for UKFast's SafeDNS"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["ukfast.net"]


def provider_parser(subparser):
    """Configure provider parser for SafeDNS"""
    subparser.description = """
        SafeDNS provider requires an API key in all interactions.
        You can generate one for your account on the following URL:
        https://my.ukfast.co.uk/applications/index.php"""
    subparser.add_argument(
        "--auth-token", help="specify the API key to authenticate with"
    )


class Provider(BaseProvider):
    """Provider SafeDNS implementation of Lexicon Provider interface."""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.ukfast.io/safedns/v1"

    def _authenticate(self):
        try:
            self._get(f"/zones/{self.domain}")
            self.domain_id = self.domain
        except requests.exceptions.HTTPError as err:
            if err.response.status_code == 404:
                raise AuthenticationError("No domain found")
            raise err

    # List all records. Return an empty list if no records found.
    # type, name and content are used to filter records.
    def _list_records(self, rtype=None, name=None, content=None):
        url = f"/zones/{self.domain_id}/records"
        records = []
        payload = {}

        # Handle pagination and iteratively collect all records
        next_url = url
        while next_url is not None:
            payload = self._get(next_url)
            if (
                "meta" in payload
                and "pagination" in payload["meta"]
                and "links" in payload["meta"]["pagination"]
                and "next" in payload["meta"]["pagination"]["links"]
            ):
                next_url = payload["meta"]["pagination"]["links"]["next"]
            else:
                next_url = None

            # Assign the returned attributes to the keys that lexicon expects
            for record in payload["data"]:
                record = self._clean_TXT_record(record)
                processed_record = {
                    "id": record["id"],
                    "name": record["name"],
                    "type": record["type"],
                    "content": record["content"],
                    "updated_at": record["updated_at"],
                    "ttl": record["ttl"],
                    "priority": record["priority"],
                }
                records.append(processed_record)

        # This is filtering logic to return only the record which matches what has
        # been passed in to the method
        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record
                for record in records
                # DNS should not be case-sensitive, so perform lower-case comparison
                if record["name"].lower() == self._full_name(name).lower()
            ]
        if content:
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", records)
        return records

    def _create_record(self, rtype, name, content):

        # Check whether the record already exists with the same  rtype, name & content.
        # If so, claim to have added the record, but dont't do anything.
        records = self._list_records(rtype, name, content)
        if records:
            LOGGER.debug(
                "create_record: (ignored, duplicate record): %s", records[0]["id"]
            )
            return True

        # Make sure TXT records are wrapped in quotes
        if content:
            content = self._add_quotes(rtype, content)

        data = {"name": self._full_name(name), "type": rtype, "content": content}

        self._post(f"/zones/{self.domain}/records", data)
        LOGGER.debug("create_record: %s", True)
        return True

    def _update_record(self, identifier, rtype=None, name=None, content=None):

        # Make sure the update won't cause a duplicate entry. If it will, fail silently
        records = self._list_records(rtype, name, content)

        if not identifier:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]
            elif len(records) > 1:
                identifier = records[0]["id"]
                LOGGER.warning(
                    "Warning, multiple records found for given parameters, "
                    "only first one will be updated: %s",
                    records,
                )
            else:
                raise Exception("Record identifier could not be found")

        # Make sure TXT records are wrapped in quotes
        if content:
            content = self._add_quotes(rtype, content)

        data = {}
        if name:
            data["name"] = self._full_name(name)
        if rtype:
            data["type"] = rtype
        if content:
            data["content"] = content

        self._patch(f"/zones/{self.domain}/records/{identifier}", data)

        LOGGER.debug("update_record: %s", True)
        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):

        delete_record_ids = []

        # If we've not been given an identifier, search for matching records.
        # NOTE, this could cause multiple records to be removed.
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_ids = [record["id"] for record in records]
        else:
            delete_record_ids.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_ids)

        for delete_record_id in delete_record_ids:
            self._delete(f"/zones/{self.domain}/records/{delete_record_id}")

        LOGGER.debug("delete_record: %s", True)
        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"{self._get_provider_option('auth_token')}",
        }
        if not url.startswith(self.api_endpoint):
            url = self.api_endpoint + url

        response = requests.request(
            action,
            url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )

        # Sort this out to work properly, writing errors to the correct place
        if "content" in response:
            resp = json.loads(response.content)
            if "errors" in resp:
                for error in resp["errors"]:
                    print("ERROR: " + error["detail"])

        # If the request fails for any reason, throw an error.
        response.raise_for_status()

        # There is no JSON returned when calling with DELETE
        if action != "DELETE":
            return response.json()

        return True

    # The content of TXT entries must be quoted. This static method ensures that.
    @staticmethod
    def _add_quotes(rtype, content):
        if rtype == "TXT":
            if not content.startswith('"') and not content.endswith('"'):
                return f'"{content}"'
        return content



"""Module provider for Godaddy"""
import hashlib
import json
import logging
from typing import Any, Dict, List, Optional, Tuple

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry  # type: ignore

from lexicon.exceptions import LexiconError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["godaddy.com", "domaincontrol.com"]


def provider_parser(subparser):
    """Generate a subparser for Godaddy"""
    subparser.add_argument("--auth-key", help="specify the key to access the API")
    subparser.add_argument("--auth-secret", help="specify the secret to access the API")


class Provider(BaseProvider):
    """
    Implements the DNS GoDaddy provider.
    Some general remarks about this provider, because it uses a weirdly designed API.
    Indeed, there is no direct way to insert, update or delete a specific record.
    Furthermore, there is no unique identifier for a record.
    Instead GoDaddy use a replace approach: for a given set of records one
    can replace this set with a new set sent through API.
    For the sake of simplicity and consistency across the provider edit methods,
    the set will be always all records in the DNS zone.
    With this approach:
        - adding a record consists in appending a record to the obtained set and call
          replace with the updated set,
        - updating a record consists in modifying a record in the obtained set and call
          replace with the updated set,
        - deleting a record consists in removing a record in the obtained set and call
          replace with the updated set.
    In parallel, as said before, there is no unique identifier.
    This provider then implement a pseudo-identifier, to allow an easy update or delete
    using the '--identifier' lexicon parameter.
    But you need to call the 'list' command just before executing and update/delete action,
    because identifier value is tied to the content of the record, and will change anytime
    something is changed in the record.
    """

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.godaddy.com/v1"

    def _authenticate(self):
        domain = self.domain

        result = self._get(f"/domains/{domain}")
        self.domain_id = result["domainId"]

    def _list_records(self, rtype=None, name=None, content=None):
        domain = self.domain

        url = f"/domains/{domain}/records"
        if rtype:
            url += f"/{rtype}"
        if name:
            url += f"/{self._relative_name(name)}"

        raws = self._get(url)

        records = []
        for raw in raws:
            records.append(
                {
                    "id": Provider._identifier(raw),
                    "type": raw["type"],
                    "name": self._full_name(raw["name"]),
                    "ttl": raw["ttl"],
                    "content": raw["data"],
                }
            )

        if content:
            records = [record for record in records if record["data"] == content]

        LOGGER.debug("list_records: %s", records)

        return records

    def _create_record(self, rtype, name, content):
        domain = self.domain
        relative_name = self._relative_name(name)
        ttl = self._get_lexicon_option("ttl")

        # Retrieve existing data in DNS zone.
        records = self._get(f"/domains/{domain}/records/{rtype}/{relative_name}")

        # Check if a record already matches given parameters
        for record in records:
            if record["data"] == content:
                LOGGER.debug(
                    "create_record (ignored, duplicate): %s %s %s", rtype, name, content
                )
                return True

        # Append a new entry corresponding to given parameters.
        data = {"type": rtype, "name": relative_name, "data": content}
        if ttl:
            data["ttl"] = int(ttl)

        records.append(data)

        # Insert the record
        self._put(f"/domains/{domain}/records/{rtype}/{relative_name}", records)

        LOGGER.debug("create_record: %s %s %s", rtype, name, content)

        return True

    def _find_matching_records(
        self,
        identifier: Optional[str],
        rtype: Optional[str],
        name: Optional[str],
        content: Optional[str],
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        # Retrieve existing data in DNS zone.
        records = self._get(f"/domains/{self.domain}/records")

        # Find matching record, either by identifier matching, or rtype + name + content matching
        if identifier:
            matching_records = [
                record
                for record in records
                if Provider._identifier(record) == identifier
            ]

            if not matching_records:
                raise LexiconError(
                    f"Could not find record matching identifier: {identifier}"
                )
        else:
            matching_records = records.copy()

            if rtype:
                matching_records = [
                    record for record in matching_records if record["type"] == rtype
                ]

            if name:
                matching_records = [
                    record
                    for record in matching_records
                    if self._relative_name(record["name"]) == self._relative_name(name)
                ]

            if content:
                matching_records = [
                    record for record in matching_records if record["data"] == content
                ]

            if not matching_records:
                suffix = f", content: {content}" if content else ""
                raise LexiconError(
                    f"Could not find record matching type: {rtype}, name: {name}{suffix}"
                )

        return matching_records, records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        # No identifier is used with GoDaddy.
        # We can rely either:
        #   - only on rtype/name to get the relevant records, both of them are required
        #     or we will could update to much records ...,
        #   - or by the pseudo-identifier provided
        # Furthermore for rtype/name approach, we cannot update all matching records, as it
        # would lead o an error (two entries of same rtype + name cannot have the same content).
        # So for rtype/name approach, we search first matching record for rtype/name on which
        # content is different, and we update it before synchronizing the DNS zone.
        if not identifier and not rtype:
            raise Exception("ERROR: rtype is required")
        if not identifier and not name:
            raise Exception("ERROR: name is required")

        matching_records, records = self._find_matching_records(
            identifier, rtype, name, None
        )

        if len(matching_records) > 1:
            LOGGER.warn(
                "Warning, multiple matching updatable records found, first one is picked."
            )

        matching_record = matching_records[0]

        # Ensure all content to update is defined
        rtype = rtype if rtype else matching_record["type"]
        name = name if name else matching_record["name"]
        content = content if content else matching_record["data"]
        relative_name = self._relative_name(name)

        # Filter out DNS zone records to focus on the target rtype
        records = [record for record in records if record["type"] == rtype]

        # Prepare update in-place
        matching_record["type"] = rtype
        matching_record["name"] = relative_name
        matching_record["data"] = content

        # Actual update, with two possible strategies
        if self._relative_name(matching_record["name"]) == relative_name:
            # If the name of the record stays the same, we use the scoped API on this name to
            # redefine specifically the records of this rtype and name, including the updated one.
            records = [
                record
                for record in records
                if self._relative_name(record["name"]) == relative_name
            ]
            self._put(
                f"/domains/{self.domain}/records/{rtype}/{relative_name}",
                records,
            )
        else:
            # If the name of the record changes, we redefine the whole set to records of this rtype
            # using the list of records, including the updated one.
            self._put(f"/domains/{self.domain}/records/{rtype}", records)

        LOGGER.debug("update_record: %s %s %s", rtype, name, content)

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        # Get the list of records to evict and the list of all current records in the DNS zone.
        matching_records, records = self._find_matching_records(
            identifier, rtype, name, content
        )

        if identifier:
            # When identifier is used, by definition only one unique record can match.
            rtype = matching_records[0]["type"]
            identifiers = [identifier]
        else:
            identifiers = [self._identifier(record) for record in matching_records]

        # Clean up the records list:
        # - by removing all records that are not matching the target type
        # - by removing all records that must be evicted in the current call
        records = [
            record
            for record in records
            if record["type"] == rtype and self._identifier(record) not in identifiers
        ]

        # Resynchronize the current set of records for the target type using the cleaned list,
        # which effectively deletes the records to evict.
        if records:
            self._put(f"/domains/{self.domain}/records/{rtype}", records)
        else:
            # List records is empty, the intention here is to delete all records of a given type.
            # Sadly GoDaddy API ignores empty arrays in a PUT request, so it is not possible to
            # do that directly using the endpoint `PUT /domains/{domain}/records/{rtype}`.
            # The trick here is to use PUT with one unique record remaining (matching_records[0]),
            # then use the endpoint `DELETE /domains/{domain}/records/{rtype}/{name}` to remove
            # the remaining record.
            self._put(f"/domains/{self.domain}/records/{rtype}", [matching_records[0]])
            self._delete(
                f"/domains/{self.domain}/records/{rtype}/{matching_records[0]['name']}"
            )

        LOGGER.debug("delete_records: %s %s %s", rtype, name, content)

        return True

    # GoDaddy provides no identifier for a record, which is a problem
    # where identifiers can be used (delete and update).
    # To circumvent this, we implement a pseudo-identifier,which is basically
    # a hash of type+name+content of a given record.
    # It is far from perfect, as the identifier will change each time
    # we change something in the record ...
    # But at least, one can use 'lexicon godaddy list ...' then
    # 'lexicon godaddy update --identifier ...' to modify specific record.
    # However, 'lexicon godaddy list ...' should be called each time DNS
    # zone had been changed to calculate new identifiers.
    @staticmethod
    def _identifier(record):
        sha256 = hashlib.sha256()
        sha256.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
        sha256.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
        sha256.update(("data=" + record.get("data", "") + ",").encode("utf-8"))
        return sha256.hexdigest()[0:7]

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # When editing DNS zone, API is unavailable for few seconds
        # (until modifications are propagated).
        # In this case, call to API will return 409 HTTP error.
        # We use the Retry extension to retry the requests until
        # we get a processable response (402 HTTP status, or an HTTP error != 409)
        try:
            retries = Retry(
                total=10,
                backoff_factor=0.5,
                status_forcelist=[409],
                allowed_methods=frozenset(["GET", "PUT", "POST", "DELETE", "PATCH"]),
            )
        except TypeError:
            # Support for urllib3<1.26
            retries = Retry(
                total=10,
                backoff_factor=0.5,
                status_forcelist=[409],
                method_whitelist=frozenset(["GET", "PUT", "POST", "DELETE", "PATCH"]),
            )

        session = requests.Session()
        session.mount("https://", HTTPAdapter(max_retries=retries))

        result = session.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data) if data else None,
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
                # GoDaddy uses a key/secret pair to authenticate
                "Authorization": f"sso-key {self._get_provider_option('auth_key')}:{self._get_provider_option('auth_secret')}",
            },
        )

        result.raise_for_status()

        try:
            # Return the JSON body response if exists.
            return result.json()
        except ValueError:
            # For some requests command (eg. PUT), GoDaddy will not
            # return any JSON, just an HTTP status without body.
            return None



"""
Module provider for EUserv

Author: Matthias Schoettle (@mschoettle), 2019

EUserv API Docs: https://support.euserv.com/api-doc/
"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["euserv.com"]

# Success response code
RC_SUCCESS = 100

API_ENDPOINT = "https://support.euserv.com/"

# Product group ID for domains
PRODUCT_ID_DOMAIN = 1


def provider_parser(subparser):
    """Configure provider parser for Euserv"""
    subparser.add_argument(
        "--auth-username", help="specify email address for authentication"
    )
    subparser.add_argument(
        "--auth-password", help="specify password for authentication"
    )


class Provider(BaseProvider):
    """Provider class for Euserv"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.api_endpoint = API_ENDPOINT
        # Order ID for the domain
        self.order_id = None
        self.domain_id = None
        self.session_id = None

    def _authenticate(self):
        if not (
            self._get_provider_option("auth_username")
            and self._get_provider_option("auth_password")
        ):
            raise ValueError(
                "username and password must be specified, add --help for details"
            )

        # Get a session ID first.
        LOGGER.info("Getting Session ID...")
        response = self._get()
        self.session_id = response["result"]["sess_id"]["value"]

        LOGGER.info("Logging in...")
        auth_response = self._get(
            query_params={
                "subaction": "login",
                "email": self._get_provider_option("auth_username"),
                "password": self._get_provider_option("auth_password"),
            }
        )

        # Find the contract number of the given domain
        orders = auth_response["result"]["orders"]

        for order in orders:
            if int(order["pg_id"]["value"]) == PRODUCT_ID_DOMAIN:
                # The description contains the description of the product itself
                # and in a second line the domain name
                order_description = order["ord_description"]["value"].split("\n")

                if order_description[1] == self.domain:
                    self.order_id = order["ord_no"]["value"]
                    break
        else:
            raise AuthenticationError("Order for domain not found")

        # Select the order for the given domain so we can use the DNS actions
        LOGGER.info("Choosing order %s", self.order_id)
        self._get(query_params={"subaction": "choose_order", "ord_no": self.order_id})

        # Retrieve domain ID
        LOGGER.info("Retrieving DNS records to find domain id for %s...", self.domain)
        domains = self._get(query_params={"subaction": "kc2_domain_dns_get_records"})

        for domain in domains["result"]["domains"]:
            if domain["dom_domain"]["value"] == self.domain:
                self.domain_id = domain["dom_id"]["value"]
                break
        else:
            raise AuthenticationError("Domain not found in DNS records")

    def _create_record(self, rtype, name, content):
        records = self._list_records(rtype, name, content)

        if records:
            LOGGER.debug("record already exists: %s", records[0])
            return True

        query_params = {
            "subaction": "kc2_domain_dns_set",
            "dom_id": self.domain_id,
            "type": rtype,
            "content": content,
        }

        if name:
            query_params["subdomain"] = self._subdomain_name(name)

        self._add_ttl(query_params)
        self._add_priority(query_params)

        response = self._get(query_params=query_params)
        LOGGER.debug("create_record response: %s", response)

        return True

    def _list_records(self, rtype=None, name=None, content=None):
        LOGGER.info(
            "Listing records for type=%s, name=%s, content=%s", rtype, name, content
        )

        query_params = {
            "subaction": "kc2_domain_dns_get_records",
            "dns_records_load_only_for_dom_id": self.domain_id,
        }

        if rtype:
            query_params["dns_records_load_type"] = rtype

        if name:
            query_params["dns_records_load_subdomain"] = self._subdomain_name(name)

        if content:
            query_params["dns_records_load_content"] = content

        payload = self._get(query_params=query_params)

        response = payload["result"]["domains"][0]

        LOGGER.debug("list_records raw response: %s", response)

        records = []

        if "dns_records" in response:
            for record in response["dns_records"]:
                processed_record = {
                    "type": record["type"]["value"],
                    "name": record["name"]["value"],
                    "ttl": record["ttl"]["value"],
                    "content": record["content"]["value"],
                    "id": record["id"]["value"],
                    "priority": record["prio"]["value"],
                }

                records.append(processed_record)

        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is None:
            LOGGER.info("No identifier for record provided, trying to find it...")
            identifier = self._find_record_identifier(rtype, name, None)

        query_params = {"subaction": "kc2_domain_dns_set", "dns_record_id": identifier}

        if rtype:
            query_params["type"] = rtype

        if name:
            query_params["subdomain"] = self._subdomain_name(name)

        if content:
            query_params["content"] = content

        self._add_ttl(query_params)
        self._add_priority(query_params)

        response = self._get(query_params=query_params)
        LOGGER.debug("update_record: %s", response)

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        record_ids = []

        if identifier is None:
            LOGGER.info("No identifier for record provided, trying to find it...")
            records = self._list_records(rtype, name, content)
            record_ids = [record["id"] for record in records]
        else:
            record_ids.append(identifier)

        LOGGER.debug("record IDs to be deleted: %s", record_ids)

        for record_id in record_ids:
            query_params = {
                "subaction": "kc2_domain_dns_remove",
                "dns_record_id": record_id,
            }

            response = response = self._get(query_params=query_params)
            LOGGER.debug("delete_record: %s", response)

        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        query_params["method"] = "json"

        if self.session_id:
            query_params["sess_id"] = self.session_id

        response = requests.request(
            action,
            self.api_endpoint,
            params=query_params,
            data=json.dumps(data),
            headers={"Content-Type": "application/json; charset=utf-8"},
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()

        response_json = response.json()

        if int(response_json["code"]) != RC_SUCCESS:
            raise Exception(
                f"Error {response_json['code']} in request: {response_json['message']}"
            )

        return response_json

    # Adds TTL parameter if passed as argument to lexicon.
    def _add_ttl(self, data):
        if self._get_lexicon_option("ttl"):
            data["ttl"] = int(self._get_lexicon_option("ttl"))

    # Adds priority parameter if passed as argument to lexicon.
    def _add_priority(self, data):
        if self._get_lexicon_option("priority"):
            data["prio"] = int(self._get_lexicon_option("priority"))

    # Find identifier of a record with the given properties.
    def _find_record_identifier(self, rtype, name, content):
        records = self._list_records(rtype, name, content)

        LOGGER.debug("records: %s", records)

        if len(records) == 1:
            return records[0]["id"]

        raise Exception("No record identifier found")

    # Get the subdomain name only for the given name.
    # This provider automatically suffixes the name with the domain name.
    def _subdomain_name(self, name):
        subdomain = self._full_name(name)
        domain_suffix = "." + self.domain

        # Remove domain name since it will be automatically added by the provider
        if subdomain.endswith(domain_suffix):
            subdomain = subdomain[: len(subdomain) - len(domain_suffix)]

        return subdomain



"""Module provider for Easyname DNS"""
import json
import logging

from bs4 import BeautifulSoup, Tag  # type: ignore
from requests import Response, Session

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["easyname.eu"]


def provider_parser(subparser):
    """Configure provider parser for Easyname DNS"""
    subparser.description = """A provider for Easyname DNS."""
    subparser.add_argument(
        "--auth-username", help="Specify username used to authenticate"
    )
    subparser.add_argument(
        "--auth-password",
        help="Specify password used to authenticate",
    )


class Provider(BaseProvider):
    """
    easyname provider
    """

    URLS = {
        "login": "https://my.easyname.com/en/login",
        "login_endpoint": "https://my.easyname.com/en/authentication-api/login",
        "domain_list": "https://my.easyname.com/domains/",
        "dashboard": "https://my.easyname.com/en/dashboard",
        "overview": "https://my.easyname.com/hosting/view-user.php",
        "dns": "https://my.easyname.com/en/domain/dns/index/domain/{}/",
        "dns_create_entry": "https://my.easyname.com/en/domain/dns/create/domain/{}",
        "dns_update_entry": "https://my.easyname.com/en/domain/dns/edit/domain/{}/id/{}",
        "dns_delete_entry": "https://my.easyname.com/en/domain/dns/delete/domain/{}/id/{}",
        "dns_delete_entry_confirm": "https://my.easyname.com/en/domain/dns/delete/domain/{}/id/{}/confirm/1",
    }

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.session = Session()
        self.domain_id = None
        self._records = None

    def _get_csrf_token(self):
        """Return the CSRF Token of easyname login form."""
        home_response = self.session.get(self.URLS["login"])
        self._log("Home", home_response)
        assert home_response.status_code == 200, "Could not load Easyname login page."

        csrf_token_field = home_response.cookies["CSRF-TOKEN"]

        assert csrf_token_field is not None, "Could not find login token."
        return csrf_token_field

    def _login(self, csrf_token):
        """Attempt to login session on easyname."""
        payload = {
            "emailAddress": self._get_provider_option("auth_username"),
            "password": self._get_provider_option("auth_password"),
        }
        headers = self.session.headers
        headers["X-CSRF-TOKEN"] = csrf_token

        login_response = self.session.post(
            self.URLS["login_endpoint"],
            data=json.dumps(payload),
            headers=headers,
            allow_redirects=True,
        )
        self._log("Login", login_response)
        json_response = json.loads(login_response.content)
        assert (
            login_response.status_code == 200
        ), "Could not login due to a network error."

        assert json_response["redirectUrl"] == self.URLS["dashboard"], (
            "Easyname login failed, bad EASYNAME_USER or EASYNAME_PASS.%s"
            % login_response.url
        )

    def _authenticate(self):

        csrf_token = self._get_csrf_token()
        self._login(csrf_token)

        domain_text_element = self._get_domain_text_of_authoritative_zone()
        self.domain_id = self._get_domain_id(domain_text_element)
        LOGGER.debug("Easyname domain ID: %s", self.domain_id)

        return True

    def _create_record(self, rtype, name, content):
        return self._create_record_internal(rtype=rtype, name=name, content=content)

    def _create_record_internal(self, rtype, name, content, identifier=None):
        """
        Create a new DNS entry in the domain zone if it does not already exist.

        Args:
          rtype (str): The DNS type (e.g. A, TXT, MX, etc) of the new entry.
          name (str): The name of the new DNS entry, e.g the domain for which a
                      MX entry shall be valid.
          content (str): The content of the new DNS entry, e.g. the mail server
                         hostname for a MX entry.
          [identifier] (str): The easyname id of a DNS entry. Use to overwrite an
                    existing entry.

        Returns:
          bool: True if the record was created successfully, False otherwise.
        """
        name = self._relative_name(name) if name is not None else name
        LOGGER.debug("Creating record with name %s", name)
        if self._is_duplicate_record(rtype, name, content):
            return True

        data = self._get_post_data_to_create_dns_entry(rtype, name, content, identifier)
        if not identifier:
            LOGGER.debug("Create DNS data: %s", data)
            create_response = self.session.post(
                self.URLS["dns_create_entry"].format(self.domain_id), data=data
            )
            self._log("Create DNS entry", create_response)
        else:
            LOGGER.debug("Update DNS data: %s", data)
            update_response = self.session.post(
                self.URLS["dns_update_entry"].format(self.domain_id, identifier),
                data=data,
            )
            self._log("Update DNS entry", update_response)

        self._invalidate_records_cache()

        # Pull a list of records and check for ours
        was_success = len(self._list_records(rtype, name, content)) > 0
        if identifier != "" and was_success:
            msg = "Successfully updated record {}".format(name)
        elif was_success:
            msg = "Successfully added record {}".format(name)
        else:
            msg = "Failed to add/update record {}".format(name)
        LOGGER.info(msg, name)
        return was_success

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """
        Delete one or more DNS entries in the domain zone that match the given
        criteria.

        Args:
          [identifier] (str): An ID to match against DNS entry easyname IDs.
          [rtype] (str): A DNS rtype (e.g. A, TXT, MX, etc) to match against DNS
                      entry types.
          [name] (str): A name to match against DNS entry names.
          [content] (str): A content to match against a DNS entry contents.

        Returns:
          bool: True if the record(s) were deleted successfully, False
                otherwise.
        """
        record_ids = self._get_matching_dns_entry_ids(identifier, rtype, name, content)
        LOGGER.debug("Record IDs to delete: %s", record_ids)

        success = True
        for rec_id in record_ids:
            delete_response = self.session.get(
                self.URLS["dns_delete_entry"].format(self.domain_id, rec_id)
            )
            delete_response_confirm = self.session.post(
                self.URLS["dns_delete_entry_confirm"].format(self.domain_id, rec_id)
            )
            self._invalidate_records_cache()
            self._log("Delete DNS entry {rec_id}", delete_response)
            # success = success and delete_response_confirm.url == success_url
            success = "feedback-message--success" in delete_response_confirm.text

        return success

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        """
        Update a DNS entry identified by identifier or name in the domain zone.
        Any non given argument will leave the current value of the DNS entry.

        Args:
          identifier (str): The easyname id of the DNS entry to update.
          [rtype] (str): The DNS rtype (e.g. A, TXT, MX, etc) of the new entry.
          [name] (str): The name of the new DNS entry, e.g the domain for which
                        a MX entry shall be valid.
          [content] (str): The content of the new DNS entry, e.g. the mail
                           server hostname for a MX entry.

        Returns:
          bool: True if the record was updated successfully, False otherwise.

        Raises:
          AssertionError: When a request returns unexpected or unknown data.
        """
        if identifier is not None:
            identifier = int(identifier)
            records = self._list_records_internal(identifier=identifier)
        else:
            records = self._list_records_internal(name=name, rtype=rtype)
        LOGGER.debug("Records to update (%d): %s", len(records), records)
        assert records, "No record found to update"
        success = True

        for record in records:
            name = name if name is not None else record["name"]
            rtype = rtype if rtype is not None else record["type"]
            content = content if content is not None else record["content"]
            success = success and self._create_record_internal(
                rtype, name, content, record["id"]
            )
        return success

    def _list_records(self, rtype=None, name=None, content=None):
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(
        self, rtype=None, name=None, content=None, identifier=None
    ):
        """
        Filter and list DNS entries of domain zone on Easyname.
        Easyname shows each entry in a HTML table row and each attribute on a
        table column.

        Args:
          [rtype] (str): Filter by DNS rtype (e.g. A, TXT, MX, etc)
          [name] (str): Filter by the name of the DNS entry, e.g the domain for
                      which a MX entry shall be valid.
          [content] (str): Filter by the content of the DNS entry, e.g. the
                           mail server hostname for a MX entry.
          [identifier] (str): Filter by the easyname id of the DNS entry.

        Returns:
          list: A list of DNS entries. A DNS entry is an object with DNS
                attribute names as keys (e.g. name, content, priority, etc)
                and additionally an id.

        Raises:
          AssertionError: When a request returns unexpected or unknown data.
        """
        name = self._full_name(name) if name is not None else name
        if self._records is None:
            records = []
            # skip the first record which contains the table header
            rows = self._get_dns_entry_trs()[1:]

            for row in rows:
                self._log("DNS list entry", row)
                try:
                    rec = {}
                    columns = row.find_all("td")
                    remove_span_from_name = columns[0].text.replace(
                        '<span style="color: var(--color-grey-500);">', ""
                    )
                    columns[0].replace_with(remove_span_from_name)
                    rec["name"] = (remove_span_from_name or "").strip()
                    rec["type"] = (columns[1].contents[1].text or "").strip()
                    rec["content"] = (columns[2].contents[1].string or "").strip()
                    rec["priority"] = (columns[3].contents[1].string or "").strip()
                    rec["ttl"] = (columns[4].contents[1].string or "").strip()
                    rec["id"] = ""
                    for a in columns[5].findAll(
                        "a", class_="button button--transparent"
                    ):
                        rec["id"] = int(a["href"].rsplit("/", 1)[-1])
                    if rec["priority"]:
                        rec["priority"] = int(rec["priority"])
                    if rec["ttl"]:
                        rec["ttl"] = int(rec["ttl"])
                except Exception as error:
                    errmsg = "Cannot parse DNS entry ({}).".format(error)
                    LOGGER.warning(errmsg)
                    raise AssertionError(errmsg)
                records.append(rec)
            self._records = records
        records = self._filter_records(self._records, rtype, name, content, identifier)
        LOGGER.debug("Final records (%d): %s", len(records), records)
        return records

    def _request(self, action="GET", url="/", data=None, query_params=None):
        pass

    def _invalidate_records_cache(self):
        """
        Invalidate DNS entries cache such that list_records will do a new
        request to retrieve DNS entries.
        """
        self._records = None

    def _get_post_data_to_create_dns_entry(self, rtype, name, content, identifier=None):
        """
        Build and return the post date that is needed to create a DNS entry.
        """
        is_update = identifier is not None
        record = None
        if is_update:
            records = self._list_records_internal(identifier=identifier)
            assert len(records) == 1, "ID is not unique or does not exist"
            record = records[0]
            LOGGER.debug("Create post data to update record: %s", record)

        data = {
            "id": str(identifier) if is_update else "",
            "name": name,
            "type": rtype,
            "content": content,
            "priority": str(record["priority"]) if is_update else "10",
            "ttl": str(record["ttl"]) if is_update else "360",
        }
        ttl = self._get_lexicon_option("ttl")
        if ttl and ttl > 360:
            data["ttl"] = str(ttl)

        prio = self._get_lexicon_option("priority")
        if prio and prio > 0:
            data["priority"] = str(prio)

        return data

    def _is_duplicate_record(self, rtype, name, content):
        """Check if DNS entry already exists."""
        records = self._list_records(rtype, name, content)
        is_duplicate = len(records) >= 1
        if is_duplicate:
            LOGGER.info("Duplicate record %s %s %s, NOOP", rtype, name, content)
        return is_duplicate

    def _get_matching_dns_entry_ids(
        self, identifier=None, rtype=None, name=None, content=None
    ):
        """Return a list of DNS entries that match the given criteria."""
        record_ids = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            record_ids = [record["id"] for record in records]
        else:
            record_ids.append(identifier)
        return record_ids

    def _get_dns_entry_trs(self):
        """
        Return the TR elements holding the DNS entries.
        """
        dns_list_response = self.session.get(self.URLS["dns"].format(self.domain_id))
        self._log("DNS list", dns_list_response)
        assert dns_list_response.status_code == 200, "Could not load DNS entries."

        html = BeautifulSoup(dns_list_response.content, "html.parser")
        self._log("DNS list", html)
        dns_table = html.find("table")
        assert dns_table is not None, "Could not find DNS entry table"

        def _is_zone_tr(elm):
            return elm.name.lower() == "tr" and (elm.has_attr("class"))

        rows = dns_table.findAll(_is_zone_tr)
        assert rows is not None and rows, "Could not find any DNS entries"
        return rows

    def _filter_records(
        self, records, rtype=None, name=None, content=None, identifier=None
    ):
        """
        Filter dns entries based on type, name or content.
        """
        if not records:
            return []
        if identifier is not None:
            LOGGER.debug("Filtering %d records by id: %s", len(records), identifier)
            records = [record for record in records if record["id"] == identifier]
        if rtype is not None:
            LOGGER.debug("Filtering %d records by type: %s", len(records), rtype)
            records = [record for record in records if record["type"] == rtype]
        if name is not None:
            LOGGER.debug("Filtering %d records by name: %s", len(records), name)
            if name.endswith("."):
                name = name[:-1]
            if rtype.lower() != "txt":
                name = name.split(".")[0]
                records = [
                    record for record in records if name == record["name"].split(".")[0]
                ]
            else:
                records = [record for record in records if name == record["name"]]
        if content is not None:
            LOGGER.debug(
                "Filtering %d records by content: %s", len(records), content.lower()
            )
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]
        return records

    def _get_domain_text_of_authoritative_zone(self):
        """Get the authoritative name zone."""
        # We are logged in, so get the domain list
        zones_response = self.session.get(self.URLS["domain_list"])
        self._log("Zone", zones_response)
        assert (
            zones_response.status_code == 200
        ), "Could not retrieve domain list due to a network error."

        html = BeautifulSoup(zones_response.content, "html.parser")
        self._log("Zone", html)
        domain_table = html.find("table")
        assert domain_table is not None, "Could not find domain table"

        # (Sub)domains can either be managed in their own zones or by the
        # zones of their parent (sub)domains. Iterate over all subdomains
        # (starting with the deepest one) and see if there is an own zone
        # for it.
        domain = self.domain or ""
        domain_text = None
        subdomains = domain.split(".")
        while True:
            domain = ".".join(subdomains)
            LOGGER.debug("Check if %s has own zone", domain)
            domain_text = domain_table.find(string=domain)
            if domain_text is not None or len(subdomains) < 3:
                break
            subdomains.pop(0)

        # Update domain to equal the zone's domain. This is important if we are
        # handling a subdomain that has no zone of itself. If we do not do
        # this, self._relative_name will strip also a part of the subdomain
        # away.
        self.domain = domain
        assert domain_text is not None, "The domain does not exist on Easyname."
        return domain_text

    def _get_domain_id(self, domain_text_element):
        """Return the easyname id of the domain."""
        try:
            # Hierarchy: TR > TD > SPAN > Domain Text
            tr_anchor = domain_text_element.parent.parent.parent
            td_anchor = tr_anchor.find("td", {"class": "entity__field taright"})
            link = td_anchor.find("a")["href"]
            domain_id = link.rsplit("/", 1)[-1]
            return domain_id
        except Exception as error:
            errmsg = (
                "Cannot get the domain id even though the domain seems "
                "to exist (%s).",
                error,
            )
            LOGGER.warning(errmsg)
            raise AssertionError(errmsg)

    def _log(self, name, element):
        """
        Log Response and Tag elements. Do nothing if elements is none of them.
        """
        if isinstance(element, Response):
            LOGGER.debug(
                "%s response: URL=%s Code=%s", name, element.url, element.status_code
            )
        elif isinstance(element, (BeautifulSoup, Tag)):
            LOGGER.debug("%s HTML:\n%s", name, element)



"""Module provider for a localzone"""
import logging
import types
from time import localtime, strftime, time
from typing import List

from .base import Provider as BaseProvider

# localzone is an optional dependency of lexicon; do not throw an ImportError if
# the dependency is unmet.
try:
    import localzone  # type: ignore
except ImportError:
    pass


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS: List[str] = []


# Monkeypatch localzone.models.Zone._increment_serial to make it compatible with dnspython 2.x
def _increment_serial(self):
    next_serial = int(strftime("%Y%m%d00", localtime(time())))

    if next_serial <= self.soa.rdata.serial:
        next_serial = self.soa.rdata.serial + 1

    if hasattr(self.soa.rdata, "replace"):
        self.soa._data = self.soa._data._replace(
            rdata=self.soa.rdata.replace(serial=next_serial)
        )
    else:
        self.soa.rdata.serial = next_serial


def _patch_zone(zone):
    zone._increment_serial = types.MethodType(_increment_serial, zone)


def provider_parser(subparser):
    """Configure provider parserfor a localzone"""
    subparser.add_argument("--filename", help="specify location of zone master file")


class Provider(BaseProvider):
    """Provider class for a localzone"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.ttl = self._get_lexicon_option("ttl")
        self.domain = self._get_lexicon_option("domain")
        self.origin = self.domain + "."
        self.filename = self._get_provider_option("filename")

    def _authenticate(self):
        # Authentication is not required for localzone.
        pass

    def _create_record(self, rtype, name, content):
        """
        Create a resource record. If a record already exists with the same
        content, do nothing.
        """
        result = False
        name = self._relative_name(name)
        ttl = None

        # TODO: shoud assert that this is an int
        if self.ttl:
            ttl = self.ttl

        with localzone.manage(self.filename, self.origin, autosave=True) as zone:
            # TODO: Remove this monkeypatch once upstream Class is fixed.
            _patch_zone(zone)
            if zone.add_record(name, rtype, content, ttl=ttl):
                result = True

        LOGGER.debug("create_record: %s", result)
        return result

    def _list_records(self, rtype=None, name=None, content=None):
        """
        Return a list of records matching the supplied params. If no params are
        provided, then return all zone records. If no records are found, return
        an empty list.
        """
        if name:
            name = self._relative_name(name)
        if not rtype:
            rtype = "ANY"

        filter_query = {"rdtype": rtype, "name": name, "content": content}

        with localzone.manage(self.filename, self.origin, autosave=True) as zone:
            # TODO: Remove this monkeypatch once upstream Class is fixed.
            _patch_zone(zone)
            records = zone.find_record(**filter_query)

        result = []
        for record in records:
            rdict = {
                "type": record.rdtype,
                "name": self._full_name(record.name),
                "ttl": record.ttl,
                "content": record.content,
                "id": record.hashid,
            }

            if rdict["type"] == "TXT":
                rdict["content"] = rdict["content"].replace('"', "")

            result.append(rdict)

        LOGGER.debug("list_records: %s", result)
        return result

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        """
        Update a record. Returns `False` if no matching record is found.
        """
        result = False

        # TODO: some providers allow content-based updates without supplying an
        # ID, and therefore `identifier` is here optional. If we don't receive
        # an ID, look it up.
        if not identifier and rtype and name:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]

        if identifier and content:
            with localzone.manage(self.filename, self.origin, autosave=True) as zone:
                # TODO: Remove this monkeypatch once upstream Class is fixed.
                _patch_zone(zone)
                if zone.update_record(identifier, content):
                    result = True

        LOGGER.debug("update_record: %s", result)
        return result

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """
        Delete record(s) matching the provided params. If there is no match, do
        nothing.
        """
        ids = []

        if identifier:
            ids.append(identifier)
        elif not identifier and rtype and name:
            records = self._list_records(rtype, name, content)
            if records:
                ids = [record["id"] for record in records]

        if ids:
            LOGGER.debug("delete_records: %s", ids)
            with localzone.manage(self.filename, self.origin, autosave=True) as zone:
                # TODO: Remove this monkeypatch once upstream Class is fixed.
                _patch_zone(zone)
                for hashid in ids:
                    zone.remove_record(hashid)
                    LOGGER.debug("delete_record: %s", hashid)

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Not required
        pass



"""Module provider for Hetzner"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)
NAMESERVER_DOMAINS = ["ns.hetzner.com"]


def provider_parser(subparser):
    """Configure a provider parser for Hetzner"""
    subparser.add_argument("--auth-token", help="Specify Hetzner DNS API token")


class Provider(BaseProvider):
    """
    Implements the Hetzner DNS Provider using the service https://dns.hetzner.com.
    It neither works for their web hosting offer "konsoleH" nor for the discontinued
    "Domain Robot" which are supported by the ``lexicon.providers.hetzner_legacy`` module.
    """

    API_VERSION = "1.0"

    """Provider class for Cloudflare"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://dns.hetzner.com/api/v1"

    def _authenticate(self):
        provider = self._get_zone_by_domain(self.domain)
        self.domain_id = provider["id"]

    def _create_record(self, rtype, name, content):
        """
        Creates a DNS record, if a record with type, name and content exits, do nothing
        :returns: A json string containing the resulting record
        :rtype: str
        """

        data = {
            # hetzner needs the FQDN name if it does not belong to the managed domain itself
            "name": self._get_record_name(self.domain, name),
            "type": rtype,
            "value": content,
            "zone_id": self.domain_id,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        records = self._list_records(rtype=rtype, name=name, content=content)
        if len(records) >= 1:
            for record in records:
                LOGGER.warning(
                    "Duplicate record %s %s %s with id %s",
                    rtype,
                    name,
                    content,
                    record["id"],
                )
            return True
        self._post("/records", data)
        return True

    def _list_records(self, rtype=None, name=None, content=None):
        """
        List all records, filterable by type, name and content
        :rtype: list
        :returns: list of records, might be empty
        """
        filter_obj = {"zone_id": self.domain_id}
        payload = self._get("/records", filter_obj)
        records = map(self._hetzner_record_to_lexicon_record, payload["records"])
        filtered_records = self._filter_records(
            records, rtype, name if name is not None else None, content
        )

        return filtered_records

    def _filter_records(self, records, rtype=None, name=None, content=None):
        return [
            record
            for record in records
            if (rtype is None or record["type"] == rtype)
            and (name is None or record["name"] == self._full_name(name))
            and (content is None or record["content"] == content)
        ]

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        """
        Create or update a record.
        :rtype: bool
        :returns: Boolean indicating success
        """
        data = {
            "type": rtype,
            "name": self._get_record_name(self.domain, name),
            "value": content,
            "zone_id": self.domain_id,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")
        update_identifier = identifier
        if update_identifier is None:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                update_identifier = records[0]["id"]
            elif len(records) < 1:
                raise Exception(
                    "No records found matching type, name and content - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type, name and content - won't update"
                )
        self._put(f"/records/{update_identifier}", data)
        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """
        Delete an existing record. If record does not exist, do nothing.
        :rtype: bool
        :returns: Boolean indicating success
        """
        delete_record_ids = []
        if identifier is None:
            records = self._list_records(rtype, name, content)
            delete_record_ids = [record["id"] for record in records]
        else:
            delete_record_ids.append(identifier)

        for record_id in delete_record_ids:
            self._delete(f"/records/{record_id}")
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers={
                "Auth-API-Token": self._get_provider_option("auth_token"),
                "Content-Type": "application/json",
            },
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    def _get_zone_by_domain(self, domain):
        """
        Requests all dns zones from your Hetzner account and searches for a specific
        one to determine the ID of it
        :param domain: Name of domain for which dns zone should be searched
        :rtype: dict
        :return: The dictionary of the zone with ``domain`` in the 'name' key
        :raises Exception: If no zone was found
        :raises KeyError, ValueError: If the response is malformed
        :raises urllib.error.HttpError: If request to /zones did not return 200
        """
        filter_obj = {"name": domain}
        payload = self._get("/zones", filter_obj)
        zones = payload["zones"]
        for zone in zones:
            if zone["name"] == domain:
                return zone
        raise AuthenticationError(f"No zone was found in account matching {domain}")

    def _get_record_name(self, domain, record_name):
        """
        Get the name attribute appropriate for hetzner api. This means it's the name
        without domain name if record name ends with managed domain name else a fqdn
        :param domain: Name of domain for which dns zone should be searched
        :param record_name: The record name to convert
        :rtype: str
        :return: The record name in an appropriate format for hetzner api
        """
        if record_name.rstrip(".").endswith(domain):
            record_name = self._relative_name(record_name)
        return record_name

    @staticmethod
    def _pretty_json(data):
        return json.dumps(data, sort_keys=True, indent=4, separators=(",", ": "))

    def _hetzner_record_to_lexicon_record(self, hetzner_record):
        lexicon_record = {
            "id": hetzner_record["id"],
            "name": self._full_name(hetzner_record["name"]),
            "content": hetzner_record["value"],
            "type": hetzner_record["type"],
        }
        if "ttl" in hetzner_record:
            lexicon_record["ttl"] = hetzner_record["ttl"]
        return lexicon_record



"""Module provider for GratisDNS"""
import logging

import requests
from bs4 import BeautifulSoup  # type: ignore

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["gratisdns.dk"]


def provider_parser(subparser):
    """Return the parser for this provider"""
    subparser.add_argument(
        "--auth-username", help="specify email address for authentication"
    )
    subparser.add_argument(
        "--auth-password", help="specify password for authentication"
    )


class Provider(BaseProvider):
    """Provider class for GratisDNS"""

    # XXX: Identifiers change on updates

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://admin.gratisdns.com"
        self.cookies = {}

    def _authenticate(self):
        # Getting required cookie "ORGID"
        payload = {
            "action": "logmein",
            "login": self._get_provider_option("auth_username"),
            "password": self._get_provider_option("auth_password"),
        }
        # Cannot allow redirects, as we miss the cookie then
        response = requests.post(self.api_endpoint, data=payload, allow_redirects=False)
        response.raise_for_status()

        if "ORGID" not in response.cookies:
            raise AuthenticationError("Unexpected auth response")
        self.cookies["ORGID"] = response.cookies["ORGID"]

        # Make sure domain exists
        # domain is stored in self.domain from BaseProvider
        domains = self._list_domains()
        for domain in domains:
            if domain == self.domain:
                # Domain name is the ID
                self.domain_id = domain
                break
        else:
            raise AuthenticationError(f"Domain {self.domain} not found")

    def _list_domains(self):
        query_params = {"action": "dns_primarydns"}
        response = self._get(query_params=query_params)
        html = BeautifulSoup(response.content, "html.parser")
        # NOTE: This could be more robust, by checking more of the tree
        domains = [x.contents[0] for x in html.find_all("th", scope="row")]
        LOGGER.debug("list_domains: %s", domains)

        return domains

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        query_params = {
            "action": "dns_primary_changeDNSsetup",
            "user_domain": self.domain_id,
        }
        response = self._get(query_params=query_params)
        html = BeautifulSoup(response.content, "html.parser")
        # List of records
        all_records = []
        # DNS records are grouped by type
        dns_record_types = html.find_all("div", "dns-records")
        for dns_record in dns_record_types:
            # Find the title of this group, aka record type
            record_type = dns_record.find("h2").contents[0].strip()
            # The entries themselves follow in the table below
            # Each row is a single entry
            entry_table_rows = dns_record.find("tbody").find_all("tr")
            for row in entry_table_rows:
                cols = row.find_all("td")
                # Format is; name, content, ttl, EDIT/DELETE button
                record_name = cols[0].contents[0]
                record_content = cols[1].contents[0]
                record_ttl = cols[2].contents[0]
                # We have to pull the ID from the button link
                record_id = None
                button_link = cols[3].find("a")
                if button_link:
                    start = button_link["href"].index("&id=") + 4
                    end = button_link["href"].index("&", start)
                    record_id = button_link["href"][start:end]

                processed_record = {
                    "type": record_type,
                    "name": record_name,
                    "ttl": record_ttl,
                    "content": record_content,
                    "id": record_id,
                }
                all_records.append(processed_record)

        records = self._filter_records(
            records=all_records, rtype=rtype, name=name, content=content
        )
        LOGGER.debug("list_records: %s", records)
        return records

    # Filter a list of records based on criteria
    # NOTE: Duplicate of lexicon/providers/transip.py _filter_records
    def _filter_records(self, records, rtype=None, name=None, content=None):
        _records = []
        for record in records:
            if (
                (not rtype or record["type"] == rtype)
                and (
                    not name or self._full_name(record["name"]) == self._full_name(name)
                )
                and (not content or record["content"] == content)
            ):
                _records.append(record)
        return _records

    def _get_content_entry(self, rtype):
        if rtype == "TXT":
            return "txtdata"
        if rtype == "A":
            return "ip"
        if rtype == "AAAA":
            return "ip"
        if rtype == "CNAME":
            return "cname"
        LOGGER.error("unknown record type")
        return None

    # Create record. If record already exists with the same content, do nothing
    def _create_record(self, rtype, name, content):
        # Figure out TTL
        ttl = 43200
        if self._get_lexicon_option("ttl"):
            ttl = self._get_lexicon_option("ttl")

        content_entry = self._get_content_entry(rtype)
        if content_entry is None:
            return False

        # Prepare query params and payload
        query_params = {
            "action": "dns_primary_record_add_" + rtype.lower(),
            "user_domain": self.domain_id,
        }
        payload = {
            "action": "dns_primary_record_added_" + rtype.lower(),
            "name": self._full_name(name),
            "ttl": ttl,
            content_entry: content,
            "user_domain": self.domain_id,
        }
        self._post(query_params=query_params, data=payload)
        # response = self._post(query_params=query_params, data=payload)
        # NOTE: We shouldn't return False on duplicate?
        #       The specification seems a little wage on this.
        # html = BeautifulSoup(response.content, "html.parser")
        # error = html.find('td', 'table-danger')
        # success = (error is None)

        LOGGER.debug("create_record: %s", True)
        return True

    def _lookup_record(self, identifier):
        # Pull all the records we know
        records = self._list_records()
        # Find the one with the provided identifier
        record = next((x for x in records if x["id"] == identifier), None)
        return record

    # Create or update a record.
    def _update_record(self, identifier=None, rtype=None, name=None, content=None):
        record = None
        # Try to find, either by identifier or by searching
        if identifier:
            # Find the relevant record by id
            record = self._lookup_record(identifier)
        else:
            # Find the relevant record by rtype and name
            records = self._list_records(rtype=rtype, name=name)
            if len(records) == 1:
                record = records[0]
                identifier = record["id"]
        # If none is found, create it
        if record is None:
            return self._create_record(rtype, name, content)

        # Update ttl if provided
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")
        # Update name if provided
        if name:
            record["name"] = self._full_name(name)
        if content:
            record["content"] = content

        content_entry = self._get_content_entry(rtype)
        if content_entry is None:
            return False

        # Prepare query params and payload
        query_params = {
            "action": "dns_primary_record_edit_" + rtype.lower(),
            "user_domain": self.domain_id,
            "id": identifier,
        }
        payload = {
            "action": "dns_primary_record_update_" + rtype.lower(),
            "id": identifier,
            "name": record["name"],
            "ttl": record["ttl"],
            content_entry: record["content"],
            "user_domain": self.domain_id,
        }
        response = self._post(query_params=query_params, data=payload)
        html = BeautifulSoup(response.content, "html.parser")
        error = html.find("td", "table-danger")
        success = error is None

        LOGGER.debug("update_record: %s", success)
        return success

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        success = True
        for record_id in delete_record_id:
            # Lookup the record to determine rtype
            if rtype is None:
                record = self._lookup_record(record_id)
                rtype = record["type"]

            # Prepare query params and payload
            query_params = {
                "action": "dns_primary_delete_" + rtype.lower(),
                "user_domain": self.domain_id,
                "id": record_id,
            }
            # _get is intentional, get really is used for deletion on gratisdns
            response = self._get(query_params=query_params)
            html = BeautifulSoup(response.content, "html.parser")
            error = html.find("td", "table-danger")
            success = success and (error is None)

        LOGGER.debug("delete_record: %s", success)
        return success

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=data,
            cookies=self.cookies,
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response
